"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from datetime import datetime, timedelta
from pytz import UTC
from decimal import Decimal

from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.tests.conftest import get_db_engine
from biweeklybudget.tests.sqlhelpers import restore_mysqldump
from biweeklybudget.models.account import Account, AcctType
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.models.ofx_statement import OFXStatement
from biweeklybudget.models.transaction import Transaction
from biweeklybudget.flaskapp.notifications import NotificationsController
from biweeklybudget.biweeklypayperiod import BiweeklyPayPeriod
from biweeklybudget.utils import dtnow
from biweeklybudget.models.txn_reconcile import TxnReconcile
from selenium.webdriver.common.by import By


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestBaseTemplateNavigation(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url)

    def test_heading(self, selenium):
        heading = selenium.find_element(By.CLASS_NAME, 'navbar-brand')
        assert heading.text == 'BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element(By.ID, 'side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_nav_links(self, selenium):
        nav = selenium.find_element(By.XPATH,
                                    "//div[contains(@class, 'sidebar-nav')]/ul"
                                    )
        navlinks = []
        for li in nav.find_elements(By.XPATH, "//li/a"):
            if li.text.strip() == '':
                continue
            navlinks.append(
                (self.relurl(li.get_attribute('href')), li.text)
            )
        assert navlinks == [
            ('/', 'Home'),
            ('/payperiods', 'Pay Periods'),
            ('/accounts', 'Accounts'),
            ('/accounts/credit-payoff', 'Credit Payoffs'),
            ('/ofx', 'OFX'),
            ('/plaid-update', 'Plaid Update'),
            ('/transactions', 'Transactions'),
            ('/reconcile', 'Reconcile'),
            ('/budgets', 'Budgets'),
            ('/scheduled', 'Scheduled'),
            ('/fuel', 'Fuel Log'),
            ('/projects', 'Projects / BoM'),
            ('/help', 'Help/Docs/Code (AGPL)')
        ]


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestBaseTemplateNotifications(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url)

    def test_notifications(self, selenium):
        div = selenium.find_element(By.ID, 'notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'

    def test_stale_accounts(self, selenium):
        div = selenium.find_elements(By.XPATH,
                                     "//div[@id='notifications-row']/div/div"
                                     )[0]
        assert div.text == '2 Accounts with stale data. View Accounts.'
        a = div.find_element(By.TAG_NAME, 'a')
        assert self.relurl(a.get_attribute('href')) == '/accounts'
        assert a.text == 'View Accounts'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
class TestBaseTmplUnreconciledNotification(AcceptanceHelper):

    def test_0_clean_db(self, dump_file_path):
        # clean the database; empty schema
        restore_mysqldump(dump_file_path, get_db_engine(), with_data=False)

    def test_01_add(self, testdb):
        a = Account(
            description='First Bank Account',
            name='BankOne',
            ofx_cat_memo_to_name=True,
            ofxgetter_config_json='{"foo": "bar"}',
            vault_creds_path='secret/foo/bar/BankOne',
            acct_type=AcctType.Bank
        )
        testdb.add(a)
        a.set_balance(
            overall_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC),
            ledger=Decimal('1.0'),
            ledger_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC)
        )
        b = Budget(
            name='1Income',
            is_periodic=True,
            description='1Income',
            starting_balance=Decimal('0.0'),
            is_income=True
        )
        testdb.add(b)
        testdb.flush()
        testdb.commit()

    def test_02_notification_none(self, selenium, base_url):
        self.get(selenium, base_url)
        assert 'unreconciled-alert' not in selenium.page_source
        assert 'Unreconciled Transactions' not in selenium.page_source

    def test_03_add_ofx(self, testdb):
        acct1 = testdb.query(Account).get(1)
        stmt1 = OFXStatement(
            account=acct1,
            filename='a1.ofx',
            file_mtime=datetime(2017, 4, 10, 12, 31, 42, tzinfo=UTC),
            as_of=datetime(2017, 4, 10, 12, 31, 42, tzinfo=UTC),
            currency='USD',
            acctid='1',
            bankid='b1',
            routing_number='r1'
        )
        testdb.add(stmt1)
        testdb.add(OFXTransaction(
            account=acct1,
            statement=stmt1,
            fitid='OFX1',
            trans_type='Deposit',
            date_posted=datetime(2017, 4, 10, 12, 3, 4, tzinfo=UTC),
            amount=Decimal('-100.0'),
            name='ofx1-income'
        ))
        # matches Transaction 2
        testdb.add(OFXTransaction(
            account=acct1,
            statement=stmt1,
            fitid='OFX2',
            trans_type='Debit',
            date_posted=datetime(2017, 4, 11, 12, 3, 4, tzinfo=UTC),
            amount=Decimal('250.0'),
            name='ofx2-trans1'
        ))
        # matches Transcation 3
        testdb.add(OFXTransaction(
            account=acct1,
            statement=stmt1,
            fitid='OFX3',
            trans_type='Purchase',
            date_posted=datetime(2017, 4, 9, 12, 3, 4, tzinfo=UTC),
            amount=Decimal('-600.0'),
            name='ofx3-trans2-st1'
        ))
        testdb.flush()
        testdb.commit()

    def test_04_notification(self, selenium, base_url):
        self.baseurl = base_url
        self.get(selenium, base_url)
        div = selenium.find_element(By.CLASS_NAME, 'unreconciled-alert')
        assert div.text == '3 Unreconciled OFXTransactions.'
        assert div.get_attribute(
            'class'
        ) == 'alert alert-warning unreconciled-alert'
        a = div.find_element(By.TAG_NAME, 'a')
        assert self.relurl(a.get_attribute('href')) == '/reconcile'
        assert a.text == 'Unreconciled OFXTransactions'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestBudgetOverBalanceNotification(AcceptanceHelper):

    def test_0_update_db(self, testdb):
        b = testdb.query(Budget).get(4)
        b.current_balance = Decimal('123456.78')
        testdb.add(b)
        testdb.flush()
        testdb.commit()

    def test_1_confirm_db(self, testdb):
        b = testdb.query(Budget).get(4)
        assert b.current_balance == Decimal('123456.78')

    def test_2_confirm_pp(self, testdb):
        acct_bal = NotificationsController.budget_account_sum(testdb)
        assert acct_bal == Decimal('12889.24')
        stand_bal = NotificationsController.standing_budgets_sum(testdb)
        assert stand_bal == Decimal('132939.07')
        pp_bal = NotificationsController.pp_sum(testdb)
        assert pp_bal == Decimal('11.76')
        unrec_amt = NotificationsController.budget_account_unreconciled(testdb)
        assert unrec_amt == Decimal('-333.33')

    def test_3_notification(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url)
        div = selenium.find_elements(By.XPATH,
                                     "//div[@id='notifications-row']/div/div"
                                     )[1]
        assert div.text == 'Combined balance of all budget-funding accounts ' \
                           '($12,889.24) is less than all allocated funds ' \
                           'total of $132,617.50 ($132,939.07 standing ' \
                           'budgets; $11.76 current pay period remaining; ' \
                           '-$333.33 unreconciled)!'
        assert div.get_attribute('class') == 'alert alert-danger'
        a = div.find_elements(By.TAG_NAME, 'a')
        assert self.relurl(a[0].get_attribute('href')) == '/accounts'
        assert a[0].text == 'budget-funding accounts'
        assert self.relurl(a[1].get_attribute('href')) == '/budgets'
        assert a[1].text == 'standing budgets'
        assert self.relurl(a[2].get_attribute('href')) == '/pay_period_for'
        assert a[2].text == 'current pay period remaining'
        assert self.relurl(a[3].get_attribute('href')) == '/reconcile'
        assert a[3].text == 'unreconciled'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestPPOverBalanceNotification(AcceptanceHelper):

    def test_0_update_db(self, testdb):
        b = testdb.query(Budget).get(4)
        b.current_balance = Decimal('1617.56')
        testdb.add(b)
        acct = testdb.query(Account).get(1)
        budget = testdb.query(Budget).get(1)
        pp = BiweeklyPayPeriod.period_for_date(dtnow(), testdb)
        tdate = pp.start_date + timedelta(days=2)
        stmt1 = OFXStatement(
            account=acct,
            filename='a1.ofx',
            file_mtime=dtnow(),
            as_of=dtnow(),
            currency='USD',
            acctid='1',
            bankid='b1',
            routing_number='r1'
        )
        testdb.add(stmt1)
        o = OFXTransaction(
            account=acct,
            statement=stmt1,
            fitid='OFX8',
            trans_type='Purchase',
            date_posted=dtnow(),
            amount=Decimal('-600.0'),
            name='ofx8-trans4'
        )
        testdb.add(o)
        t = Transaction(
            date=tdate,
            budget_amounts={budget: Decimal('600.00')},
            description='trans6',
            account=acct
        )
        testdb.add(t)
        testdb.add(TxnReconcile(transaction=t, ofx_trans=o))
        t2 = Transaction(
            date=tdate,
            budget_amounts={budget: Decimal('34000.00')},
            description='transFoo',
            account=acct
        )
        testdb.add(t2)
        testdb.flush()
        testdb.commit()

    def test_1_confirm_pp(self, testdb):
        acct_bal = NotificationsController.budget_account_sum(testdb)
        assert acct_bal == Decimal('12889.24')
        stand_bal = NotificationsController.standing_budgets_sum(testdb)
        assert stand_bal == Decimal('11099.85')
        pp_bal = NotificationsController.pp_sum(testdb)
        assert pp_bal == Decimal('11.76')
        unrec_amt = NotificationsController.budget_account_unreconciled(testdb)
        assert unrec_amt == Decimal('33666.67')

    def test_2_notification(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url)
        div = selenium.find_elements(By.XPATH,
                                     "//div[@id='notifications-row']/div/div"
                                     )[1]
        assert div.text == 'Combined balance of all budget-funding accounts ' \
                           '($12,889.24) is less than all allocated funds ' \
                           'total of $44,778.28 ($11,099.85 standing ' \
                           'budgets; $11.76 current pay period remaining; ' \
                           '$33,666.67 unreconciled)!'
        assert div.get_attribute('class') == 'alert alert-danger'
        a = div.find_elements(By.TAG_NAME, 'a')
        assert self.relurl(a[0].get_attribute('href')) == '/accounts'
        assert a[0].text == 'budget-funding accounts'
        assert self.relurl(a[1].get_attribute('href')) == '/budgets'
        assert a[1].text == 'standing budgets'
        assert self.relurl(a[2].get_attribute('href')) == '/pay_period_for'
        assert a[2].text == 'current pay period remaining'
        assert self.relurl(a[3].get_attribute('href')) == '/reconcile'
        assert a[3].text == 'unreconciled'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestUnderBalanceNotification(AcceptanceHelper):

    def test_0_update_db(self, testdb):
        testdb.query(Account).get(1).set_balance(
            ledger=Decimal('428890.24'),
            avail=Decimal('428890.24'),
            ledger_date=dtnow(),
            avail_date=dtnow(),
            overall_date=dtnow()
        )
        testdb.flush()
        testdb.commit()

    def test_1_confirm_db(self, testdb):
        a = testdb.query(Account).get(1)
        assert a.balance.ledger == Decimal('428890.24')
        assert a.balance.avail == Decimal('428890.24')

    def test_2_confirm_pp(self, testdb):
        acct_bal = NotificationsController.budget_account_sum(testdb)
        assert acct_bal == Decimal('428990.47')
        stand_bal = NotificationsController.standing_budgets_sum(testdb)
        assert stand_bal == Decimal('10766.52')
        pp_bal = NotificationsController.pp_sum(testdb)
        assert pp_bal == Decimal('11.76')
        unrec_amt = NotificationsController.budget_account_unreconciled(testdb)
        assert unrec_amt == Decimal('-333.33')

    def test_3_notification(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url)
        div = selenium.find_elements(By.XPATH,
                                     "//div[@id='notifications-row']/div/div"
                                     )[1]
        assert div.text == 'Combined balance of all budget-funding accounts ' \
                           '($428,990.47) is more than all allocated funds ' \
                           'total of $10,444.95 ($10,766.52 standing ' \
                           'budgets; $11.76 current pay period remaining; ' \
                           '-$333.33 unreconciled)!'
        assert div.get_attribute('class') == 'alert alert-info'
        a = div.find_elements(By.TAG_NAME, 'a')
        assert self.relurl(a[0].get_attribute('href')) == '/accounts'
        assert a[0].text == 'budget-funding accounts'
        assert self.relurl(a[1].get_attribute('href')) == '/budgets'
        assert a[1].text == 'standing budgets'
        assert self.relurl(a[2].get_attribute('href')) == '/pay_period_for'
        assert a[2].text == 'current pay period remaining'
        assert self.relurl(a[3].get_attribute('href')) == '/reconcile'
        assert a[3].text == 'unreconciled'



"""add ScheduledTransaction sales_tax field

Revision ID: 1bb9e6a1c07c
Revises: d01774fa3ae3
Create Date: 2026-01-10 13:34:45.628528

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1bb9e6a1c07c'
down_revision = 'd01774fa3ae3'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'scheduled_transactions',
        sa.Column('sales_tax', sa.Numeric(precision=10, scale=4), nullable=False)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('scheduled_transactions', 'sales_tax')
    # ### end Alembic commands ###



"""add Transaction field to store sales tax paid

Revision ID: d01774fa3ae3
Revises: f5a002127934
Create Date: 2024-01-24 19:47:44.629591

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd01774fa3ae3'
down_revision = 'f5a002127934'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'transactions',
        sa.Column('sales_tax', sa.Numeric(precision=10, scale=4), nullable=False)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('transactions', 'sales_tax')
    # ### end Alembic commands ###



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
import json
from decimal import Decimal, ROUND_UP
from datetime import datetime, timedelta

from flask.views import MethodView
from flask import render_template, request, jsonify

from biweeklybudget.flaskapp.jsonencoder import MagicJSONEncoder
from biweeklybudget.flaskapp.app import app
from biweeklybudget.db import db_session
from biweeklybudget.interest import InterestHelper
from biweeklybudget.models.dbsetting import DBSetting
from biweeklybudget.utils import fmt_currency, dtnow
from biweeklybudget.models.account import NoInterestChargedError, Account
from biweeklybudget.models.ofx_statement import OFXStatement
from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.flaskapp.views.formhandlerview import FormHandlerView

logger = logging.getLogger(__name__)


class CreditPayoffsView(MethodView):
    """
    Render the top-level GET /accounts/credit-payoff view using
    ``credit-payoffs.html`` template.
    """

    def _payoffs_list(self, ih):
        """
        Return a payoffs list suitable for rendering.

        :param ih: interest helper instance
        :type ih: biweeklybudget.interest.InterestHelper
        :return: list of payoffs suitable for rendering
        :rtype: list
        """
        res = ih.calculate_payoffs()
        payoffs = []
        for methname in sorted(res.keys(), reverse=True):
            tmp = {
                'name': methname,
                'description': res[methname]['description'],
                'doc': res[methname]['doc'],
                'results': []
            }
            if 'error' in res[methname]:
                tmp['error'] = res[methname]['error']
                continue
            total_pymt = Decimal('0')
            total_int = Decimal('0')
            total_next = Decimal('0')
            max_mos = 0
            for k in sorted(res[methname]['results'].keys()):
                r = res[methname]['results'][k]
                acct = ih.accounts[k]
                tmp['results'].append({
                    'name': '%s (%d) (%s @ %s%%)' % (
                        acct.name, k,
                        fmt_currency(abs(acct.balance.ledger)),
                        (acct.effective_apr * Decimal('100')).quantize(
                            Decimal('.01')
                        )
                    ),
                    'total_payments': r['total_payments'],
                    'total_interest': r['total_interest'],
                    'payoff_months': r['payoff_months'],
                    'next_payment': r['next_payment']
                })
                total_pymt += r['total_payments']
                total_int += r['total_interest']
                total_next += r['next_payment']
                if r['payoff_months'] > max_mos:
                    max_mos = r['payoff_months']
            tmp['total'] = {
                'total_payments': total_pymt,
                'total_interest': total_int,
                'payoff_months': max_mos,
                'next_payment': total_next
            }
            payoffs.append(tmp)
        return payoffs

    def _payment_settings_dict(self, settings_json):
        """
        Given the JSON string payment settings, return a dict of payment
        settings as expected by :py:class:`~.InterestHelper` kwargs.

        :param settings_json: payment settings JSON
        :type settings_json: str
        :return: payment settings dict
        :rtype: dict
        """
        res = {'increases': {}, 'onetimes': {}}
        j = json.loads(settings_json)
        for i in j['increases']:
            if not i['enabled']:
                continue
            d = datetime.strptime(i['date'], '%Y-%m-%d').date()
            res['increases'][d] = Decimal(i['amount'])
        for i in j['onetimes']:
            if not i['enabled']:
                continue
            d = datetime.strptime(i['date'], '%Y-%m-%d').date()
            res['onetimes'][d] = Decimal(i['amount'])
        return res

    def get(self):
        setting = db_session.query(DBSetting).get('credit-payoff')
        if setting is None:
            pymt_settings_json = json.dumps({'increases': [], 'onetimes': []})
        else:
            pymt_settings_json = setting.value
        pymt_settings_kwargs = self._payment_settings_dict(pymt_settings_json)
        try:
            ih = InterestHelper(db_session, **pymt_settings_kwargs)
            mps = sum(ih.min_payments.values())
            payoffs = self._payoffs_list(ih)
        except NoInterestChargedError as ex:
            resp = render_template(
                'credit-payoffs-no-interest-error.html',
                acct_name=ex.account.name,
                acct_id=ex.account.id
            )
            return resp, 500
        return render_template(
            'credit-payoffs.html',
            monthly_pymt_sum=mps.quantize(Decimal('.01'), rounding=ROUND_UP),
            payoffs=payoffs,
            pymt_settings_json=pymt_settings_json
        )


class PayoffSettingsFormHandler(MethodView):
    """
    Handle POST /settings/credit-payoff
    """

    def post(self):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        data = request.get_json(force=True, silent=True)
        if data is None:
            logger.error('Error parsing request JSON')
            return jsonify({
                'success': False,
                'error_message': 'Error parsing JSON'
            })
        setting = db_session.query(DBSetting).get('credit-payoff')
        if setting is None:
            setting = DBSetting(name='credit-payoff')
            logger.info('new DBSetting name=credit-payoff')
        else:
            logger.info('Existing DBSetting name=credit-payoff value=%s',
                        setting.value)
        fixeddata = {'increases': [], 'onetimes': []}
        for key in ['increases', 'onetimes']:
            for d in sorted(data[key], key=lambda k: k['date']):
                if d['date'] == '' or d['amount'] == '':
                    continue
                fixeddata[key].append(d)
        val = json.dumps(fixeddata, sort_keys=True, cls=MagicJSONEncoder)
        logger.info('Changing setting value to: %s', val)
        setting.value = val
        db_session.add(setting)
        db_session.commit()
        return jsonify({
            'success': True,
            'success_message': 'Successfully updated setting '
                               '"credit-payoff" in database.'
        })


class AccountOfxAjax(MethodView):
    """
    Handle GET /ajax/account_ofx_ajax/<int:account_id> endpoint.
    """

    def get(self, account_id):
        res = []
        q = db_session.query(OFXStatement).filter(
            OFXStatement.account_id.__eq__(account_id)
        ).order_by(
            OFXStatement.as_of.desc(),
            OFXStatement.file_mtime.desc()
        ).all()
        for stmt in q:
            if stmt.as_of < (dtnow() - timedelta(days=32)):
                break
            res.append({
                'filename': stmt.filename,
                'as_of': stmt.as_of.strftime('%Y-%m-%d'),
                'ledger_bal': stmt.ledger_bal
            })
        return jsonify({
            'account_id': account_id,
            'statements': res
        })


class AccountOfxFormHandler(FormHandlerView):
    """
    Handle POST /forms/credit-payoff-account-ofx
    """

    def validate(self, data):
        pass

    def submit(self, data):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :param data: submitted form data
        :type data: dict
        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        acct = db_session.query(Account).get(int(data['id']))
        if acct is None:
            raise RuntimeError('ERROR: No Account with ID %s' % data['id'])
        stmt = db_session.query(OFXStatement).filter(
            OFXStatement.account_id.__eq__(acct.id),
            OFXStatement.filename.__eq__(data['filename'])
        ).one()
        if stmt is None:
            raise RuntimeError(
                'ERROR: No OFXStatement for account %d with filename %s' % (
                    acct.id, data['filename']
                )
            )
        int_amt = Decimal(data['interest_amt'])
        if int_amt < Decimal('0'):
            int_amt = int_amt * Decimal('-1')
        trans = OFXTransaction(
            account=acct,
            statement=stmt,
            fitid='%s-MANUAL-CCPAYOFF' % dtnow().strftime('%Y%m%d%H%M%S'),
            trans_type='debit',
            date_posted=stmt.as_of,
            amount=int_amt,
            name='Interest Charged - MANUALLY ENTERED',
            is_interest_charge=True
        )
        logger.info(
            'Adding manual interest transaction to OFXTransactions: '
            'account_id=%d statement_filename=%s statement=%s '
            'OFXTransaction=%s', acct.id, data['filename'], stmt,
            trans
        )
        db_session.add(trans)
        db_session.commit()
        return 'Successfully saved OFXTransaction with FITID %s in database' \
               '.' % trans.fitid


app.add_url_rule(
    '/accounts/credit-payoff',
    view_func=CreditPayoffsView.as_view('credit_payoffs_view')
)
app.add_url_rule(
    '/settings/credit-payoff',
    view_func=PayoffSettingsFormHandler.as_view('payoff_settings_form')
)
app.add_url_rule(
    '/ajax/account_ofx_ajax/<int:account_id>',
    view_func=AccountOfxAjax.as_view('account_ofx_ajax')
)
app.add_url_rule(
    '/forms/credit-payoff-account-ofx',
    view_func=AccountOfxFormHandler.as_view('payoff_account_ofx_form')
)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import argparse
import logging
import atexit
import importlib

from biweeklybudget.db import init_db, cleanup_db, db_session, engine
from biweeklybudget.cliutils import set_log_debug, set_log_info
from biweeklybudget.models.base import Base

logger = logging.getLogger(__name__)


def parse_args():
    default_mod = 'biweeklybudget.tests.fixtures.sampledata'
    default_cls = 'SampleDataLoader'
    p = argparse.ArgumentParser(description='Load initial data to DB')
    p.add_argument('-v', '--verbose', dest='verbose', action='count', default=0,
                   help='verbose output. specify twice for debug-level output.')
    p.add_argument('-m', '--data-module', dest='modname', action='store',
                   type=str, default=default_mod,
                   help='data fixture module (default: %s)' % default_mod)
    p.add_argument('-c', '--data-class', dest='clsname', action='store',
                   type=str, default=default_cls,
                   help='data fixture class (default: %s)' % default_cls)
    args = p.parse_args()
    return args


def main():
    global logger
    logging.basicConfig(
        level=logging.WARNING,
        format="[%(asctime)s %(levelname)s] %(message)s"
    )
    logger = logging.getLogger()

    args = parse_args()

    # set logging level
    if args.verbose > 1:
        set_log_debug(logger)
    elif args.verbose == 1:
        set_log_info(logger)

    atexit.register(cleanup_db)
    Base.metadata.reflect(engine)
    Base.metadata.drop_all(engine)
    db_session.flush()
    db_session.commit()
    init_db()

    logger.info('Loading data from: %s.%s', args.modname, args.clsname)
    klass = getattr(importlib.import_module(args.modname), args.clsname)
    inst = klass(db_session)
    logger.info('Loading data')
    inst.load()
    db_session.commit()
    logger.info('Data loaded.')


if __name__ == "__main__":
    main()



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from flask.views import MethodView
from flask import render_template, jsonify
from datetime import datetime
from decimal import Decimal

from biweeklybudget.flaskapp.app import app
from biweeklybudget.db import db_session
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.models.budget_transaction import BudgetTransaction
from biweeklybudget.flaskapp.views.formhandlerview import FormHandlerView
from biweeklybudget.models.account import Account
from biweeklybudget.models.utils import do_budget_transfer
from biweeklybudget.biweeklypayperiod import BiweeklyPayPeriod
from biweeklybudget.models.transaction import Transaction
from biweeklybudget.utils import dtnow

logger = logging.getLogger(__name__)


class BudgetsView(MethodView):
    """
    Render the GET /budgets view using the ``budgets.html`` template.
    """

    def get(self):
        standing = db_session.query(Budget).filter(
            Budget.is_periodic.__eq__(False)
        ).order_by(Budget.name).all()
        periodic = db_session.query(Budget).filter(
            Budget.is_periodic.__eq__(True)
        ).order_by(Budget.name).all()
        accts = {a.name: a.id for a in db_session.query(Account).all()}
        budgets = {}
        active_budgets = {}
        for b in db_session.query(Budget).all():
            k = b.name
            if b.is_income:
                k = '%s (i)' % b.name
            budgets[b.id] = k
            if b.is_active:
                active_budgets[b.id] = k
        return render_template(
            'budgets.html',
            standing=standing,
            periodic=periodic,
            accts=accts,
            budgets=budgets,
            active_budgets=active_budgets
        )


class OneBudgetView(MethodView):
    """
    Render the GET /budgets/<int:budget_id> view using the ``budgets.html``
    template.
    """

    def get(self, budget_id):
        standing = db_session.query(Budget).filter(
            Budget.is_periodic.__eq__(False)
        ).order_by(Budget.name).all()
        periodic = db_session.query(Budget).filter(
            Budget.is_periodic.__eq__(True)
        ).order_by(Budget.name).all()
        accts = {a.name: a.id for a in db_session.query(Account).all()}
        budgets = {}
        active_budgets = {}
        for b in db_session.query(Budget).all():
            k = b.name
            if b.is_income:
                k = '%s (i)' % b.name
            budgets[b.id] = k
            if b.is_active:
                active_budgets[b.id] = k
        return render_template(
            'budgets.html',
            standing=standing,
            periodic=periodic,
            budget_id=budget_id,
            accts=accts,
            budgets=budgets,
            active_budgets=active_budgets
        )


class BudgetAjax(MethodView):
    """
    Handle GET /ajax/budget/<int:budget_id> endpoint.
    """

    def get(self, budget_id):
        budget = db_session.query(Budget).get(budget_id)
        return jsonify(budget.as_dict)


class BudgetFormHandler(FormHandlerView):
    """
    Handle POST /forms/budget
    """

    def validate(self, data):
        """
        Validate the form data. Return None if it is valid, or else a hash of
        field names to list of error strings for each field.

        :param data: submitted form data
        :type data: dict
        :return: None if no errors, or hash of field name to errors for that
          field
        """
        have_errors = False
        errors = {k: [] for k in data.keys()}
        if data.get('name', '').strip() == '':
            errors['name'].append('Name cannot be empty')
            have_errors = True
        if (
            data['is_periodic'] is True and
            data['starting_balance'].strip() == ''
        ):
            errors['starting_balances'].append(
                'Starting balance must be specified for periodic budgets.'
            )
            have_errors = True
        if (
            data['is_periodic'] is False and
            data['current_balance'].strip() == ''
        ):
            errors['current_balances'].append(
                'Current balance must be specified for standing budgets.'
            )
            have_errors = True
        if have_errors:
            return errors
        return None

    def submit(self, data):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :param data: submitted form data
        :type data: dict
        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        if 'id' in data and data['id'].strip() != '':
            # updating an existing budget
            budget = db_session.query(Budget).get(int(data['id']))
            if budget is None:
                raise RuntimeError("Error: no Budget with ID %s" % data['id'])
            action = 'updating Budget ' + data['id']
        else:
            budget = Budget()
            action = 'creating new Budget'
        budget.name = data['name'].strip()
        budget.description = data['description'].strip()
        budget.is_periodic = data['is_periodic']
        if data['is_periodic'] is True:
            budget.starting_balance = Decimal(data['starting_balance'])
        else:
            budget.current_balance = Decimal(data['current_balance'])
        budget.is_active = data['is_active']
        budget.is_income = data['is_income']
        budget.omit_from_graphs = data['omit_from_graphs']
        logger.info('%s: %s', action, budget.as_dict)
        db_session.add(budget)
        db_session.commit()
        return 'Successfully saved Budget %d in database.' % budget.id


class BudgetTxfrFormHandler(FormHandlerView):
    """
    Handle POST /forms/budget_transfer
    """

    def validate(self, data):
        """
        Validate the form data. Return None if it is valid, or else a hash of
        field names to list of error strings for each field.

        :param data: submitted form data
        :type data: dict
        :return: None if no errors, or hash of field name to errors for that
          field
        """
        have_errors = False
        errors = {k: [] for k in data.keys()}
        if data['date'].strip() == '':
            errors['date'].append('Transactions must have a date')
            have_errors = True
        else:
            try:
                datetime.strptime(data['date'], '%Y-%m-%d').date()
            except Exception:
                errors['date'].append(
                    'Date "%s" is not valid (YYYY-MM-DD)' % data['date']
                )
                have_errors = True
        if float(data['amount']) == 0:
            errors['amount'].append('Amount cannot be zero')
            have_errors = True
        if float(data['amount']) < 0:
            errors['amount'].append('Amount cannot be negative')
            have_errors = True
        if data['account'] == 'None':
            errors['account'].append('Transactions must have an account')
            have_errors = True
        if data['from_budget'] == 'None':
            errors['from_budget'].append('from_budget cannot be empty')
            have_errors = True
        to_budget = db_session.query(Budget).get(int(data['to_budget']))
        if to_budget is None:
            errors['to_budget'].append('to_budget ID does not exist')
            have_errors = True
        if data['to_budget'] == 'None':
            errors['to_budget'].append('to_budget cannot be empty')
            have_errors = True
        if have_errors:
            return errors
        return None

    def submit(self, data):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :param data: submitted form data
        :type data: dict
        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        # get the data
        trans_date = datetime.strptime(data['date'], '%Y-%m-%d').date()
        amt = Decimal(data['amount'])
        acct = db_session.query(Account).get(int(data['account']))
        if acct is None:
            raise RuntimeError(
                "Error: no Account with ID %s" % data['account']
            )
        from_budget = db_session.query(Budget).get(int(data['from_budget']))
        if from_budget is None:
            raise RuntimeError(
                "Error: no Budget with ID %s" % data['from_budget']
            )
        to_budget = db_session.query(Budget).get(int(data['to_budget']))
        if to_budget is None:
            raise RuntimeError(
                "Error: no Budget with ID %s" % data['to_budget']
            )
        notes = data['notes'].strip()
        desc = 'Budget Transfer - %s from %s (%d) to %s (%d)' % (
            amt, from_budget.name, from_budget.id, to_budget.name,
            to_budget.id
        )
        logger.info(desc)
        res = do_budget_transfer(db_session, trans_date, amt, acct,
                                 from_budget, to_budget, notes=notes)
        db_session.commit()
        return 'Successfully saved Transactions %d and %d in database.' % (
            res[0].id, res[1].id
        )


class BudgetSpendingChartView(MethodView):
    """
    Handle GET /ajax/chart-data/budget-spending/<str:aggregation> endpoint.
    """

    def get(self, aggregation):
        if aggregation == 'by-pay-period':
            return self._by_pay_period()
        elif aggregation == 'by-month':
            return self._by_month()
        raise RuntimeError('Unknown aggregation type: %s' % aggregation)

    def _by_pay_period(self):
        min_txn = db_session.query(Transaction).order_by(
            Transaction.date.asc()
        ).first()
        budget_names = self._budget_names()
        logger.debug('budget_names=%s', budget_names)
        records = []
        budgets_present = set()
        pp = BiweeklyPayPeriod.period_for_date(min_txn.date, db_session)
        dt_now = dtnow().date()
        while pp.end_date <= dt_now:
            sums = pp.budget_sums
            logger.debug('sums=%s', sums)
            records.append({
                budget_names[y]: sums[y]['spent'] for y in sums.keys()
                if y in budget_names
            })
            records[-1]['date'] = pp.start_date.strftime('%Y-%m-%d')
            budgets_present.update(
                [budget_names[y] for y in sums.keys() if y in budget_names]
            )
            pp = pp.next
        res = {
            'data': records,
            'keys': sorted(list(budgets_present))
        }
        return jsonify(res)

    def _by_month(self):
        dt_now = dtnow().date()
        budget_names = self._budget_names()
        logger.debug('budget_names=%s', budget_names)
        records = {}
        budgets_present = set()
        for t in db_session.query(Transaction).filter(
            Transaction.budget_transactions.any(
                BudgetTransaction.budget.has(is_active=True)
            ),
            Transaction.date.__le__(dt_now)
        ).all():
            for bt in t.budget_transactions:
                if bt.budget_id not in budget_names:
                    continue
                budg_name = bt.budget.name
                budgets_present.add(budg_name)
                ds = t.date.strftime('%Y-%m')
                if ds not in records:
                    records[ds] = {'date': ds}
                if budg_name not in records[ds]:
                    records[ds][budg_name] = Decimal('0')
                records[ds][budg_name] += bt.amount
        result = [records[k] for k in sorted(records.keys())]
        res = {
            'data': result,
            'keys': sorted(list(budgets_present))
        }
        return jsonify(res)

    def _budget_names(self):
        return {
            x.id: x.name for x in db_session.query(Budget).filter(
                Budget.is_income.__eq__(False),
                Budget.omit_from_graphs.__eq__(False)
            ).all()
        }


app.add_url_rule('/budgets', view_func=BudgetsView.as_view('budgets_view'))
app.add_url_rule(
    '/budgets/<int:budget_id>',
    view_func=OneBudgetView.as_view('one_budget_view')
)
app.add_url_rule(
    '/ajax/budget/<int:budget_id>',
    view_func=BudgetAjax.as_view('budget_ajax')
)
app.add_url_rule(
    '/forms/budget',
    view_func=BudgetFormHandler.as_view('budget_form')
)
app.add_url_rule(
    '/forms/budget_transfer',
    view_func=BudgetTxfrFormHandler.as_view('budget_transfer_form')
)
app.add_url_rule(
    '/ajax/chart-data/budget-spending/<string:aggregation>',
    view_func=BudgetSpendingChartView.as_view('budget_spending_chart_view')
)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
import logging
from decimal import Decimal
from datetime import date
from sqlalchemy import text

from biweeklybudget.tests.migrations.migration_test_helpers import MigrationTest

logger = logging.getLogger(__name__)


@pytest.mark.migrations
class TestAddBudgetTransactionModel(MigrationTest):
    """
    Test for revision 04e61490804b
    """

    migration_rev = '04e61490804b'

    def data_setup(self, engine):
        """method to setup sample data in empty tables"""
        sql = [
            "INSERT INTO accounts SET name='acct1', acct_type=1, "
            "reconcile_trans=0;",
            "INSERT INTO accounts SET name='acct2', acct_type=2, "
            "reconcile_trans=1;",
            "INSERT INTO budgets SET name='budg1';",
            "INSERT INTO budgets SET name='budg2';",
            "INSERT INTO scheduled_transactions SET amount=567.34, "
            "description='st1', account_id=2, budget_id=1, is_active=1, "
            "num_per_period=1;",
            "INSERT INTO transactions SET description='t1', account_id=1, "
            "budget_id=1, actual_amount=123.45;",
            "INSERT INTO transactions SET description='t2', account_id=1, "
            "budget_id=2, actual_amount=345.67, budgeted_amount=567.89;",
            "INSERT INTO transactions SET description='t3', account_id=2, "
            "budget_id=1, actual_amount=345.19, budgeted_amount=567.34, "
            "date='2018-02-24', notes='myNotes', scheduled_trans_id=1;",
        ]
        conn = engine.connect()
        for s in sql:
            logger.debug('Executing: %s', s)
            conn.execute(text(s))
        conn.commit()
        conn.close()

    def verify_before(self, engine):
        """method to verify data before forward migration, and after reverse"""
        conn = engine.connect()
        txns = [
            dict(r._mapping) for r in conn.execute(text('SELECT * FROM transactions'))
        ]
        conn.close()
        assert txns == [
            {
                'id': 1,
                'date': None,
                'actual_amount': Decimal('123.4500'),
                'budgeted_amount': None,
                'description': 't1',
                'notes': None,
                'account_id': 1,
                'scheduled_trans_id': None,
                'budget_id': 1,
                'transfer_id': None
            },
            {
                'id': 2,
                'date': None,
                'actual_amount': Decimal('345.67'),
                'budgeted_amount': Decimal('567.89'),
                'description': 't2',
                'notes': None,
                'account_id': 1,
                'scheduled_trans_id': None,
                'budget_id': 2,
                'transfer_id': None
            },
            {
                'id': 3,
                'date': date(2018, 2, 24),
                'actual_amount': Decimal('345.19'),
                'budgeted_amount': Decimal('567.34'),
                'description': 't3',
                'notes': 'myNotes',
                'account_id': 2,
                'scheduled_trans_id': 1,
                'budget_id': 1,
                'transfer_id': None
            }
        ]

    def verify_after(self, engine):
        """method to verify data after forward migration"""
        conn = engine.connect()
        txns = [
            dict(r._mapping) for r in conn.execute(text('SELECT * FROM transactions'))
        ]
        bts = [
            dict(r._mapping) for r in conn.execute(text('SELECT * FROM budget_transactions'))
        ]
        conn.close()
        assert txns == [
            {
                'id': 1,
                'date': None,
                'budgeted_amount': None,
                'description': 't1',
                'notes': None,
                'account_id': 1,
                'scheduled_trans_id': None,
                'transfer_id': None,
                'planned_budget_id': None
            },
            {
                'id': 2,
                'date': None,
                'budgeted_amount': Decimal('567.89'),
                'description': 't2',
                'notes': None,
                'account_id': 1,
                'scheduled_trans_id': None,
                'planned_budget_id': 2,
                'transfer_id': None
            },
            {
                'id': 3,
                'date': date(2018, 2, 24),
                'budgeted_amount': Decimal('567.34'),
                'description': 't3',
                'notes': 'myNotes',
                'account_id': 2,
                'scheduled_trans_id': 1,
                'planned_budget_id': 1,
                'transfer_id': None
            }
        ]
        assert bts == [
            {
                'id': 1,
                'trans_id': 1,
                'amount': Decimal('123.4500'),
                'budget_id': 1
            },
            {
                'id': 2,
                'trans_id': 2,
                'amount': Decimal('345.6700'),
                'budget_id': 2
            },
            {
                'id': 3,
                'trans_id': 3,
                'amount': Decimal('345.19'),
                'budget_id': 1
            }
        ]



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from sqlalchemy import (
    Column, Integer, String, Boolean, Date, SmallInteger, Numeric,
    ForeignKey, func
)
from sqlalchemy.orm import relationship, validates
from sqlalchemy.sql import case

from biweeklybudget.models.base import Base, ModelAsDict
from biweeklybudget.utils import date_suffix
from sqlalchemy.ext.hybrid import hybrid_property

#: Weekday names indexed by day_of_week value (0=Monday, 6=Sunday)
WEEKDAY_NAMES = ['Monday', 'Tuesday', 'Wednesday', 'Thursday',
                 'Friday', 'Saturday', 'Sunday']

#: Month abbreviations indexed by month number (1-12, index 0 unused)
MONTH_ABBREVS = ['', 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']


class ScheduledTransaction(Base, ModelAsDict):

    __tablename__ = 'scheduled_transactions'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)

    #: Amount of the transaction
    amount = Column(Numeric(precision=10, scale=4), nullable=False)

    #: Sales tax paid on this transaction
    sales_tax = Column(
        Numeric(precision=10, scale=4), nullable=False, default=0.0
    )

    #: description
    description = Column(String(254), nullable=False, index=True)

    #: notes
    notes = Column(String(254))

    #: ID of the account the transaction is against
    account_id = Column(Integer, ForeignKey('accounts.id'))

    #: Relationship - :py:class:`~.Account` the transaction is against
    account = relationship(
        "Account", backref="scheduled_transactions"
    )

    #: ID of the budget the transaction is against
    budget_id = Column(Integer, ForeignKey('budgets.id'))

    #: Relationship - :py:class:`~.Budget` the transaction is against
    budget = relationship(
        "Budget", backref="scheduled_transactions"
    )

    #: whether the scheduled transaction is enabled or disabled
    is_active = Column(Boolean, default=True)

    #: Denotes a scheduled transaction that will happen once on the given date
    date = Column(Date)

    #: Denotes a scheduled transaction that happens on the same day of each
    #: month
    day_of_month = Column(SmallInteger)

    #: Denotes a scheduled transaction that happens N times per pay period
    num_per_period = Column(SmallInteger)

    #: Denotes a scheduled transaction that happens every week on a given day
    #: (0=Monday, 6=Sunday)
    day_of_week = Column(SmallInteger)

    #: Month for annual scheduled transaction (1-12)
    annual_month = Column(SmallInteger)

    #: Day for annual scheduled transaction (1-31)
    annual_day = Column(SmallInteger)

    def __repr__(self):
        return "<ScheduledTransaction(id=%d)>" % (
            self.id
        )

    @validates('day_of_month')
    def validate_day_of_month(self, _, value):
        if value is None:
            return value
        assert value > 0
        assert value <= 28
        return value

    @validates('num_per_period')
    def validate_num_per_period(self, _, value):
        if value is None:
            return value
        assert value > 0
        return value

    @validates('day_of_week')
    def validate_day_of_week(self, _, value):
        if value is None:
            return value
        assert value >= 0
        assert value <= 6
        return value

    @validates('annual_month')
    def validate_annual_month(self, _, value):
        if value is None:
            return value
        assert value >= 1
        assert value <= 12
        return value

    @validates('annual_day')
    def validate_annual_day(self, _, value):
        if value is None:
            return value
        assert value >= 1
        assert value <= 31
        return value

    @hybrid_property
    def recurrence_str(self):
        """
        Return a string describing the recurrence interval. This is a string
        of the format ``YYYY-mm-dd``, ``N per period``, ``N(st|nd|rd|th)``,
        ``Every {Weekday}``, or ``{Mon} {Day}{suffix}`` where ``N`` is an
        integer.

        :return: string describing recurrence interval
        :rtype: str
        """
        if self.schedule_type == 'date':
            return self.date.strftime('%Y-%m-%d')
        if self.schedule_type == 'monthly':
            return '%d%s' % (
                self.day_of_month,
                date_suffix(self.day_of_month)
            )
        if self.schedule_type == 'per period':
            return '%d per period' % self.num_per_period
        if self.schedule_type == 'weekly':
            return 'Every %s' % WEEKDAY_NAMES[self.day_of_week]
        if self.schedule_type == 'annual':
            return '%s %d%s' % (
                MONTH_ABBREVS[self.annual_month],
                self.annual_day,
                date_suffix(self.annual_day)
            )
        return None

    @recurrence_str.expression
    def recurrence_str(cls):
        # SQLAlchemy 2.0: case() no longer takes a list, pass tuples directly
        return case(
            (
                cls.date.isnot(None),
                func.date_format(cls.date, '%Y-%m-%d')
            ),
            (
                cls.day_of_month.isnot(None),
                cls.day_of_month
            ),
            (
                cls.num_per_period.isnot(None),
                func.concat(cls.num_per_period, ' per period')
            ),
            (
                cls.day_of_week.isnot(None),
                func.concat(
                    'Every ',
                    func.elt(
                        cls.day_of_week + 1,
                        'Monday', 'Tuesday', 'Wednesday', 'Thursday',
                        'Friday', 'Saturday', 'Sunday'
                    )
                )
            ),
            (
                cls.annual_month.isnot(None),
                func.concat(
                    func.elt(
                        cls.annual_month,
                        'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                        'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'
                    ),
                    ' ',
                    cls.annual_day
                )
            ),
            else_=''
        )

    @hybrid_property
    def schedule_type(self):
        """
        Return a string describing the type of schedule; one of ``date`` (a
        specific Date), ``per period`` (a number per pay period), ``monthly``
        (a given day of the month), ``weekly`` (a given day of the week), or
        ``annual`` (a given month and day each year).

        :return: string describing type of schedule
        :rtype: str
        """
        if self.date is not None:
            return 'date'
        if self.day_of_month is not None:
            return 'monthly'
        if self.num_per_period is not None:
            return 'per period'
        if self.day_of_week is not None:
            return 'weekly'
        if self.annual_month is not None:
            return 'annual'
        return None

    @schedule_type.expression
    def schedule_type(cls):
        return case(
            (cls.date.isnot(None), 'date'),
            (cls.day_of_month.isnot(None), 'monthly'),
            (cls.num_per_period.isnot(None), 'per period'),
            (cls.day_of_week.isnot(None), 'weekly'),
            (cls.annual_month.isnot(None), 'annual'),
            else_=''
        )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from sqlalchemy import (
    Column, Integer, String, Boolean, Date, ForeignKey, SmallInteger, Numeric,
    desc, inspect
)
from decimal import Decimal, ROUND_FLOOR
from sqlalchemy.orm import relationship, validates
from biweeklybudget.models.base import Base, ModelAsDict
from biweeklybudget.utils import dtnow

logger = logging.getLogger(__name__)


class Vehicle(Base, ModelAsDict):

    __tablename__ = 'vehicles'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)

    #: Name of vehicle
    name = Column(String(254))

    #: whether active or historical
    is_active = Column(Boolean, default=True)

    def __repr__(self):
        return "<Vehicle(id=%s, name=%s)>" % (
            self.id, self.name
        )


class FuelFill(Base, ModelAsDict):

    __tablename__ = 'fuellog'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)

    #: date of the fill
    date = Column(Date, default=dtnow().date())

    #: ID of the vehicle
    vehicle_id = Column(Integer, ForeignKey('vehicles.id'))

    #: The vehicle
    vehicle = relationship(
        "Vehicle", backref="fuellog", uselist=False
    )

    #: Odometer reading of the vehicle, in miles
    odometer_miles = Column(Integer)

    #: Number of miles the vehicle thinks it's traveled since the last fill.
    reported_miles = Column(SmallInteger)

    #: Number of miles actually traveled since the last fill.
    calculated_miles = Column(SmallInteger)

    #: Fuel level before fill, as a percentage (Integer 0-100)
    level_before = Column(SmallInteger)

    #: Fuel level after fill, as a percentage (Integer 0-100)
    level_after = Column(SmallInteger)

    #: Location of fill - usually a gas station name/address
    fill_location = Column(String(254))

    #: Fuel cost per gallon
    cost_per_gallon = Column(Numeric(precision=10, scale=4))

    #: Total cost of fill
    total_cost = Column(Numeric(precision=10, scale=4))

    #: Total amount of fuel (gallons)
    gallons = Column(Numeric(precision=10, scale=4))

    #: MPG as reported by the vehicle itself
    reported_mpg = Column(Numeric(precision=10, scale=4))

    #: Calculated MPG, based on last fill
    calculated_mpg = Column(Numeric(precision=10, scale=4))

    #: Notes
    notes = Column(String(254))

    def __repr__(self):
        return "<FuelFill(id=%s, vehicle=%s, date=%s)>" % (
            self.id, self.vehicle_id, self.date
        )

    @validates('gallons')
    def validate_gallons(self, _, value):
        assert value > 0
        return value

    @validates('odometer_miles')
    def validate_odometer_miles(self, _, value):
        prev = self._previous_entry()
        if prev is None:
            logger.warning(
                'Previous fill is None; cannot validate odometer_miles'
            )
            return value
        assert self.odometer_miles > prev.odometer_miles
        return value

    def _previous_entry(self):
        """
        Get the previous fill for this vehicle by odometer reading, or None.

        :return: the previous fill for this vehicle, by odometer reading, or
          None.
        :rtype: biweeklybudget.models.fuel.FuelFill
        """
        if self.vehicle is None:
            logger.warning(
                'vehicle is None; cannot obtain previous fill for %s', self
            )
            return None
        if self.odometer_miles is None:
            logger.warning(
                'odometer_miles is None; cannot obtain previous fill for %s',
                self
            )
            return None
        return inspect(self).session.query(FuelFill).filter(
            FuelFill.vehicle.__eq__(self.vehicle),
            FuelFill.odometer_miles.__lt__(self.odometer_miles)
        ).order_by(
            desc(FuelFill.odometer_miles)
        ).first()

    def calculate_mpg(self):
        """
        Calculate ``calculated_mpg`` field.

        :returns: True if recalculate, False if unable to calculate
        :rtype: bool
        """
        if self.gallons is None:
            logger.warning(
                'Gallons is none; cannot recalculate MPG for %s', self
            )
            return False
        if self.odometer_miles is None:
            logger.warning(
                'odometer_miles is none; cannot recalculate MPG for %s', self
            )
            return False
        prev = self._previous_entry()
        if prev is None:
            logger.warning('Previous entry is None; cannot recalculate MPG '
                           'for %s', self)
            return False
        distance = self.odometer_miles - prev.odometer_miles
        self.calculated_miles = distance
        self.calculated_mpg = (
            (distance * Decimal(1.0)) / self.gallons
        ).quantize(Decimal('.001'), rounding=ROUND_FLOOR)
        logger.debug('Calculate MPG for fill %d: distance=%s mpg=%s',
                     self.id, distance, self.calculated_mpg)
        inspect(self).session.add(self)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from biweeklybudget.models import *
from biweeklybudget.models.account import AcctType
from biweeklybudget.utils import dtnow
from biweeklybudget.flaskapp.jsonencoder import MagicJSONEncoder
from datetime import timedelta, datetime
from pytz import UTC
from decimal import Decimal
import json


class SampleDataLoader(object):

    def __init__(self, db_session):
        self.db = db_session
        self.accounts = {}
        self.statements = {}
        self.budgets = {}
        self.scheduled_transactions = []
        self.transactions = {}
        self.plaid_items = {}
        self.plaid_accounts = {}
        self.dt = dtnow()

    def load(self):
        self._fuellog()
        self.db.flush()
        self.db.commit()
        for f in self.db.query(FuelFill).all():
            f.calculate_mpg()
        self.db.flush()
        self.db.commit()
        self.db.add(
            DBSetting(
                name='prime_rate',
                value=json.dumps({
                    'value': '0.0050',
                    'date': dtnow()
                }, cls=MagicJSONEncoder)
            )
        )
        self.plaid_items = self._plaid_items()
        self.plaid_accounts = self._plaid_accounts()
        self.db.flush()
        self.db.commit()
        self._projects()
        self.accounts = {
            'BankOne': self._bank_one(),
            'BankTwoStale': self._bank_two_stale(),
            'CreditOne': self._credit_one(),
            'CreditTwo': self._credit_two(),
            'InvestmentOne': self._investment_one(),
            'DisabledBank': self._disabled_bank()
        }
        self.budgets = self._budgets()
        self.scheduled_transactions = self._scheduled_transactions()
        self.transactions = self._transactions()
        self.db.add(TxnReconcile(
            ofx_trans=self.accounts['BankOne']['transactions'][0][1],
            transaction=self.transactions[0],
            note='reconcile notes',
            reconciled_at=datetime(2017, 4, 10, 8, 9, 11, tzinfo=UTC)
        ))
        self.db.flush()
        self.db.commit()

    def _plaid_items(self):
        res = {
            'PlaidItem1': PlaidItem(
                item_id='PlaidItem1',
                access_token='AccessToken1',
                institution_name='Inst1',
                last_updated=self.dt
            ),
            'PlaidItem2': PlaidItem(
                item_id='PlaidItem2',
                access_token='AccessToken2',
                institution_name='Inst2',
                last_updated=self.dt
            ),
        }
        for item in res.values():
            self.db.add(item)
        return res

    def _plaid_accounts(self):
        res = {
            'PlaidAcct1': PlaidAccount(
                item_id='PlaidItem1',
                account_id='PlaidAcct1',
                name='Acct1',
                mask='foo',
                account_type='depository',
                account_subtype='checking'
            ),
            'PlaidAcct2': PlaidAccount(
                item_id='PlaidItem1',
                account_id='PlaidAcct2',
                name='Acct2',
                mask='foo',
                account_type='credit',
                account_subtype='credit card'
            ),
            'PlaidAcct3': PlaidAccount(
                item_id='PlaidItem2',
                account_id='PlaidAcct3',
                name='Acct3',
                mask='foo',
                account_type='investment',
                account_subtype='401k'
            ),
            'PlaidAcct4': PlaidAccount(
                item_id='PlaidItem1',
                account_id='PlaidAcct4',
                name='Acct4',
                mask='foo4',
                account_type='depository',
                account_subtype='checking'
            ),
        }
        for item in res.values():
            self.db.add(item)
        return res

    def _budgets(self):
        res = {
            'Periodic1': Budget(
                name='Periodic1',
                is_periodic=True,
                description='P1desc',
                starting_balance=Decimal(100)
            ),
            'Periodic2': Budget(
                name='Periodic2',
                is_periodic=True,
                description='P2desc',
                starting_balance=Decimal(234)
            ),
            'Periodic3 Inactive': Budget(
                name='Periodic3 Inactive',
                is_periodic=True,
                description='P3desc',
                starting_balance=Decimal('10.23'),
                is_active=False
            ),
            'Standing1': Budget(
                name='Standing1',
                is_periodic=False,
                description='S1desc',
                current_balance=Decimal('1284.23'),
                omit_from_graphs=True
            ),
            'Standing2': Budget(
                name='Standing2',
                is_periodic=False,
                description='S2desc',
                current_balance=Decimal('9482.29')
            ),
            'Standing3 Inactive': Budget(
                name='Standing3 Inactive',
                is_periodic=False,
                description='S3desc',
                current_balance=Decimal('-92.29'),
                is_active=False
            ),
            'Income': Budget(
                name='Income',
                is_periodic=True,
                description='IncomeDesc',
                starting_balance=Decimal('2345.67'),
                is_income=True,
                omit_from_graphs=True
            )
        }
        for x in [
            'Periodic1', 'Periodic2', 'Periodic3 Inactive',
            'Standing1', 'Standing2', 'Standing3 Inactive',
            'Income'
        ]:
            self.db.add(res[x])
        return res

    def _scheduled_transactions(self):
        res = [
            ScheduledTransaction(
                amount=Decimal('111.11'),
                sales_tax=Decimal('1.23'),
                description='ST1',
                notes='notesST1',
                account=self.accounts['BankOne']['account'],
                budget=self.budgets['Periodic1'],
                date=(self.dt + timedelta(days=4)).date()
            ),
            ScheduledTransaction(
                amount=Decimal('222.22'),
                sales_tax=Decimal('0.0'),
                description='ST2',
                notes='notesST2',
                account=self.accounts['BankOne']['account'],
                budget=self.budgets['Periodic2'],
                day_of_month=4
            ),
            ScheduledTransaction(
                amount=Decimal('-333.33'),
                sales_tax=Decimal('4.56'),
                description='ST3',
                notes='notesST3',
                account=self.accounts['BankTwoStale']['account'],
                budget=self.budgets['Standing1'],
                num_per_period=1
            ),
            ScheduledTransaction(
                amount=Decimal('444.44'),
                sales_tax=Decimal('0.0'),
                description='ST4',
                notes='notesST4',
                account=self.accounts['BankOne']['account'],
                budget=self.budgets['Periodic1'],
                date=(self.dt + timedelta(days=5)).date(),
                is_active=False
            ),
            ScheduledTransaction(
                amount=Decimal('555.55'),
                sales_tax=Decimal('7.89'),
                description='ST5',
                notes='notesST5',
                account=self.accounts['BankOne']['account'],
                budget=self.budgets['Periodic2'],
                day_of_month=5,
                is_active=False
            ),
            ScheduledTransaction(
                amount=Decimal('666.66'),
                sales_tax=Decimal('0.0'),
                description='ST6',
                notes='notesST6',
                account=self.accounts['BankTwoStale']['account'],
                budget=self.budgets['Standing1'],
                num_per_period=3,
                is_active=False
            )
        ]
        for x in res:
            self.db.add(x)
        return res

    def _transactions(self):
        res = [
            Transaction(
                date=(self.dt + timedelta(days=4)).date(),
                budget_amounts={self.budgets['Periodic1']: Decimal('111.13')},
                budgeted_amount=Decimal('111.11'),
                description='T1foo',
                notes='notesT1',
                account=self.accounts['BankOne']['account'],
                scheduled_trans=self.scheduled_transactions[0],
                planned_budget=self.budgets['Periodic1'],
                sales_tax=Decimal('0.0')
            ),
            Transaction(
                date=self.dt.date(),
                budget_amounts={self.budgets['Standing1']: Decimal('-333.33')},
                budgeted_amount=Decimal('-333.33'),
                description='T2',
                notes='notesT2',
                account=self.accounts['BankTwoStale']['account'],
                scheduled_trans=self.scheduled_transactions[2],
                planned_budget=self.budgets['Standing1']
            ),
            Transaction(
                date=(self.dt - timedelta(days=2)).date(),
                budget_amounts={self.budgets['Periodic2']: Decimal('222.22')},
                description='T3',
                notes='notesT3',
                account=self.accounts['CreditOne']['account'],
                sales_tax=Decimal('12.34')
            ),
            Transaction(
                date=(self.dt - timedelta(days=35)).date(),
                budget_amounts={
                    self.budgets['Periodic2']: Decimal('222.22'),
                    self.budgets['Periodic1']: Decimal('100.10')
                },
                description='T4split',
                notes='notesT4split',
                account=self.accounts['CreditOne']['account'],
                sales_tax=Decimal('34.56')
            )
        ]
        for x in res:
            self.db.add(x)
        return res

    def _add_account(self, acct, statements, transactions):
        self.db.add(acct)
        for s in statements:
            self.db.add(s)
            acct.set_balance(
                overall_date=s.as_of,
                ledger=s.ledger_bal,
                ledger_date=s.ledger_bal_as_of,
                avail=s.avail_bal,
                avail_date=s.avail_bal_as_of
            )
        for s in transactions.keys():
            for t in transactions[s]:
                self.db.add(t)
        return {
            'account': acct,
            'statements': statements,
            'transactions': transactions
        }

    def _bank_one(self):
        acct = Account(
            description='First Bank Account',
            name='BankOne',
            ofx_cat_memo_to_name=True,
            ofxgetter_config_json='{"foo": "bar"}',
            vault_creds_path='secret/foo/bar/BankOne',
            acct_type=AcctType.Bank,
            re_interest_charge='^interest-charge',
            re_interest_paid='^interest-paid',
            re_payment='^(payment|thank you)',
            re_late_fee='^Late Fee',
            re_other_fee='^re-other-fee',
            plaid_item_id='PlaidItem1',
            plaid_account_id='PlaidAcct1',
        )
        statements = [
            OFXStatement(
                account=acct,
                filename='/stmt/BankOne/0',
                file_mtime=(self.dt - timedelta(hours=46)),
                currency='USD',
                bankid='BankOne',
                routing_number='11',
                acct_type='Checking',
                acctid='1111',
                type='Bank',
                as_of=(self.dt - timedelta(hours=46)),
                ledger_bal=Decimal('12345.67'),
                ledger_bal_as_of=(self.dt - timedelta(hours=46)),
                avail_bal=Decimal('12340.00'),
                avail_bal_as_of=(self.dt - timedelta(hours=46))
            ),
            OFXStatement(
                account=acct,
                filename='/stmt/BankOne/1',
                file_mtime=(self.dt - timedelta(hours=14)),
                currency='USD',
                bankid='BankOne',
                routing_number='11',
                acct_type='Checking',
                acctid='1111',
                type='Bank',
                as_of=(self.dt - timedelta(hours=14)),
                ledger_bal=Decimal('12789.01'),
                ledger_bal_as_of=(self.dt - timedelta(hours=14)),
                avail_bal=Decimal('12563.18'),
                avail_bal_as_of=(self.dt - timedelta(hours=14))
            )
        ]
        transactions = {
            0: [
                OFXTransaction(
                    account=acct,
                    statement=statements[0],
                    fitid='BankOne.0.0',
                    trans_type='Credit',
                    date_posted=(self.dt - timedelta(days=7)),
                    amount=Decimal('1234.56'),
                    name='BankOne.0.0'
                ),
                OFXTransaction(
                    account=acct,
                    statement=statements[0],
                    fitid='BankOne.0.1',
                    trans_type='Debit',
                    date_posted=(self.dt - timedelta(days=6)),
                    amount=Decimal('-20.00'),
                    name='Late Fee'
                )
            ],
            1: []
        }
        for x in range(1, 21):
            mult = 1
            if x % 2 == 0:
                mult = -1
            amt = (11 * x) * mult
            transactions[1].append(OFXTransaction(
                    account=acct,
                    statement=statements[1],
                    fitid='BankOne.1.%d' % x,
                    trans_type='Debit',
                    date_posted=(self.dt - timedelta(days=6, hours=x)),
                    amount=Decimal(amt),
                    name='Generated Trans %d' % x
            ))
        return self._add_account(acct, statements, transactions)

    def _bank_two_stale(self):
        acct = Account(
            description='Stale Bank Account',
            name='BankTwoStale',
            ofx_cat_memo_to_name=False,
            ofxgetter_config_json='{"foo": "baz"}',
            vault_creds_path='secret/foo/bar/BankTwo',
            acct_type=AcctType.Bank,
            is_active=True,
            negate_ofx_amounts=True
        )
        statements = [
            OFXStatement(
                account=acct,
                filename='/stmt/BankTwoStale/0',
                file_mtime=(self.dt - timedelta(days=18)),
                currency='USD',
                bankid='BankTwoStale',
                routing_number='22',
                acct_type='Savings',
                acctid='2222',
                type='Bank',
                as_of=(self.dt - timedelta(days=18)),
                ledger_bal=Decimal('100.23'),
                ledger_bal_as_of=(self.dt - timedelta(days=18))
            )
        ]
        transactions = {
            0: [
                OFXTransaction(
                    account=acct,
                    statement=statements[0],
                    fitid='0',
                    trans_type='Debit',
                    date_posted=(self.dt - timedelta(days=23)),
                    amount=Decimal('432.19'),
                    name='Transfer to Other Account'
                ),
                OFXTransaction(
                    account=acct,
                    statement=statements[0],
                    fitid='1',
                    trans_type='Interest',
                    date_posted=(self.dt - timedelta(days=22)),
                    amount=Decimal('0.23'),
                    name='Interest Paid',
                    memo='Some Date'
                )
            ]
        }
        return self._add_account(acct, statements, transactions)

    def _credit_one(self):
        acct = Account(
            description='First Credit Card, limit 2000',
            name='CreditOne',
            ofx_cat_memo_to_name=False,
            acct_type=AcctType.Credit,
            credit_limit=Decimal('2000.00'),
            is_active=True,
            prime_rate_margin=Decimal('0.0050'),
            negate_ofx_amounts=True,
            interest_class_name='AdbCompoundedDaily',
            min_payment_class_name='MinPaymentAmEx',
            re_interest_charge='^INTEREST CHARGED TO',
            re_payment='.*Online Payment, thank you.*',
            re_late_fee='^Late Fee',
            re_other_fee='^re-other-fee',
            plaid_item_id='PlaidItem1',
            plaid_account_id='PlaidAcct2',
        )
        statements = [
            OFXStatement(
                account=acct,
                filename='/stmt/CreditOne/3',
                file_mtime=(self.dt - timedelta(days=1)),
                currency='USD',
                bankid='CreditOne',
                acct_type='Credit',
                acctid='CreditOneAcctId',
                type='Credit',
                as_of=(self.dt - timedelta(days=1)),
                ledger_bal=Decimal('-435.29'),
                ledger_bal_as_of=(self.dt - timedelta(days=1)),
            ),
            OFXStatement(
                account=acct,
                filename='/stmt/CreditOne/1',
                file_mtime=(self.dt - timedelta(days=30, hours=13)),
                currency='USD',
                bankid='CreditOne',
                acct_type='Credit',
                acctid='CreditOneAcctId',
                type='Credit',
                as_of=(self.dt - timedelta(days=30, hours=13)),
                ledger_bal=Decimal('-876.54'),
                ledger_bal_as_of=(self.dt - timedelta(days=30, hours=13)),
            ),
            OFXStatement(
                account=acct,
                filename='/stmt/CreditOne/0',
                file_mtime=(self.dt - timedelta(hours=13)),
                currency='USD',
                bankid='CreditOne',
                acct_type='Credit',
                acctid='CreditOneAcctId',
                type='Credit',
                as_of=(self.dt - timedelta(hours=13)),
                ledger_bal=Decimal('-952.06'),
                ledger_bal_as_of=(self.dt - timedelta(hours=13)),
            )
        ]
        transactions = {
            0: [
                OFXTransaction(
                    account=acct,
                    statement=statements[2],
                    fitid='T1',
                    trans_type='Purchase',
                    date_posted=(self.dt - timedelta(hours=22)),
                    amount=Decimal('-123.81'),
                    name='123.81 Credit Purchase T1',
                    memo='38328',
                    description='CreditOneT1Desc',
                    checknum=123,
                    mcc='T1mcc',
                    sic='T1sic'
                ),
                OFXTransaction(
                    account=acct,
                    statement=statements[2],
                    fitid='T2',
                    trans_type='credit',
                    date_posted=(self.dt - timedelta(days=2)),
                    amount=Decimal('52.00'),
                    name='$52.00 Online Payment, thank you',
                    memo='38328',
                    description='CreditOneT2Desc'
                ),
                OFXTransaction(
                    account=acct,
                    statement=statements[2],
                    fitid='T3',
                    trans_type='debit',
                    date_posted=(self.dt - timedelta(hours=13)),
                    amount=Decimal('-16.25'),
                    name='INTEREST CHARGED TO STANDARD PUR',
                    memo='38328',
                    description='CreditOneT3Desc'
                )
            ],
            1: [
                OFXTransaction(
                    account=acct,
                    statement=statements[1],
                    fitid='T2-1',
                    trans_type='credit',
                    date_posted=(self.dt - timedelta(days=32)),
                    amount=Decimal('60.00'),
                    name='$60.00 Online Payment, thank you',
                    memo='38328',
                    description='CreditOneT2Desc'
                ),
                OFXTransaction(
                    account=acct,
                    statement=statements[1],
                    fitid='T2-2',
                    trans_type='debit',
                    date_posted=(self.dt - timedelta(days=30, hours=13)),
                    amount=Decimal('-25.94'),
                    name='INTEREST CHARGED TO STANDARD PUR',
                    memo='38328',
                    description='CreditOneT3Desc'
                )
            ]
        }
        return self._add_account(acct, statements, transactions)

    def _credit_two(self):
        acct = Account(
            description='Credit 2 limit 5500',
            name='CreditTwo',
            ofx_cat_memo_to_name=False,
            ofxgetter_config_json='',
            vault_creds_path='/foo/bar',
            acct_type=AcctType.Credit,
            credit_limit=Decimal(5500),
            is_active=True,
            apr=Decimal('0.1000'),
            interest_class_name='AdbCompoundedDaily',
            min_payment_class_name='MinPaymentDiscover'
        )
        statements = [
            OFXStatement(
                account=acct,
                filename='/stmt/CreditTwo/0',
                file_mtime=(self.dt - timedelta(hours=36)),
                currency='USD',
                bankid='CreditTwo',
                acct_type='Credit',
                acctid='',
                type='CreditCard',
                as_of=(self.dt - timedelta(hours=36)),
                ledger_bal=Decimal('-5498.65'),
                ledger_bal_as_of=(self.dt - timedelta(hours=36))
            )
        ]
        transactions = {
            0: [
                OFXTransaction(
                    account=acct,
                    statement=statements[0],
                    fitid='001',
                    trans_type='Purchase',
                    date_posted=(self.dt - timedelta(hours=36)),
                    amount=Decimal('28.53'),
                    name='Interest Charged',
                    memo=''
                ),
                OFXTransaction(
                    account=acct,
                    statement=statements[0],
                    fitid='002',
                    trans_type='Credit',
                    date_posted=(self.dt - timedelta(days=5)),
                    amount=Decimal('-50.00'),
                    name='Online Payment - Thank You',
                    memo=''
                )
            ]
        }
        return self._add_account(acct, statements, transactions)

    def _investment_one(self):
        acct = Account(
            description='Investment One Stale',
            name='InvestmentOne',
            ofx_cat_memo_to_name=False,
            ofxgetter_config_json='',
            vault_creds_path='',
            acct_type=AcctType.Investment,
            is_active=True,
            reconcile_trans=False,
            plaid_item_id='PlaidItem2',
            plaid_account_id='PlaidAcct3',
        )
        statements = [
            OFXStatement(
                account=acct,
                filename='/stmt/InvOne/0',
                file_mtime=(self.dt - timedelta(days=13, hours=6)),
                currency='USD',
                acct_type='Retirement',
                brokerid='InvOneBroker',
                acctid='1000001',
                type='Investment',
                as_of=(self.dt - timedelta(days=13, hours=6)),
                ledger_bal=Decimal('10362.91'),
                ledger_bal_as_of=(self.dt - timedelta(days=13, hours=6))
            )
        ]
        return self._add_account(acct, statements, {})

    def _disabled_bank(self):
        acct = Account(
            description='Disabled Bank Account',
            name='DisabledBank',
            ofx_cat_memo_to_name=True,
            ofxgetter_config_json='{"bar": "baz"}',
            vault_creds_path='',
            acct_type=AcctType.Bank,
            is_active=False
        )
        statements = [
            OFXStatement(
                account=acct,
                filename='/stmt/DisabledBank/0',
                file_mtime=(self.dt - timedelta(days=41)),
                currency='USD',
                bankid='111DDD111',
                routing_number='111DDD111',
                acct_type='Savings',
                acctid='D1111111',
                type='Bank',
                as_of=(self.dt - timedelta(hours=46)),
                ledger_bal=Decimal('10.00'),
                ledger_bal_as_of=(self.dt - timedelta(days=41)),
                avail_bal=Decimal('10.00'),
                avail_bal_as_of=(self.dt - timedelta(days=41))
            )
        ]
        transactions = {
            0: [
                OFXTransaction(
                    account=acct,
                    statement=statements[0],
                    fitid='001',
                    trans_type='Credit',
                    date_posted=(self.dt - timedelta(days=43)),
                    amount=Decimal('0.01'),
                    name='Interest Paid',
                    memo=''
                ),
                OFXTransaction(
                    account=acct,
                    statement=statements[0],
                    fitid='002',
                    trans_type='Debit',
                    date_posted=(self.dt - timedelta(days=51)),
                    amount=Decimal('3218.87'),
                    name='ATM Withdrawal',
                    memo='Disabled002Memo',
                    description='Disabled002Desc'
                )
            ]
        }
        return self._add_account(acct, statements, transactions)

    def _example_bank(self):
        """Sample to copy"""
        acct = Account(
            description='',
            name='',
            ofx_cat_memo_to_name=False,
            ofxgetter_config_json='',
            vault_creds_path='',
            acct_type=AcctType.Bank,
            # credit_limit=0,
            is_active=True
        )
        statements = [
            OFXStatement(
                account=acct,
                filename='/stmt/BankOne/0',
                file_mtime=(self.dt - timedelta(hours=46)),
                currency='USD',
                bankid='',
                routing_number='',
                acct_type='',
                brokerid='',
                acctid='',
                type='',
                as_of=(self.dt - timedelta(hours=46)),
                ledger_bal=Decimal('12345.67'),
                ledger_bal_as_of=(self.dt - timedelta(hours=46)),
                avail_bal=Decimal('12340.00'),
                avail_bal_as_of=(self.dt - timedelta(hours=46))
            )
        ]
        transactions = {
            0: [
                OFXTransaction(
                    account=acct,
                    statement=statements[0],
                    fitid='',
                    trans_type='',
                    date_posted=(self.dt - timedelta(days=7)),
                    amount=Decimal('1234.56'),
                    name='',
                    memo=''
                ),
                OFXTransaction(
                    account=acct,
                    statement=statements[0],
                    fitid='',
                    trans_type='',
                    date_posted=(self.dt - timedelta(days=7)),
                    amount=Decimal('1234.56'),
                    name='',
                    memo=''
                )
            ]
        }
        return self._add_account(acct, statements, transactions)

    def _fuellog(self):
        v1 = Vehicle(name='Veh1')
        self.db.add(v1)
        v2 = Vehicle(name='Veh2')
        self.db.add(v2)
        v3 = Vehicle(name='Veh3Inactive', is_active=False)
        self.db.add(v3)
        start_dt = dtnow()
        for veh_num, veh in {1: v1, 2: v2, 3: v3}.items():
            veh_str = 'v%d' % veh_num
            for i in range(0, 2):
                dt = start_dt + timedelta(days=i)
                self.db.add(FuelFill(
                    date=dt.date(),
                    vehicle=veh,
                    odometer_miles=(1000 + (i * 10) + veh_num),
                    reported_miles=(100 + (i * 10) + veh_num),
                    level_before=(i * 10) + veh_num,
                    level_after=(100 - (i * 10) - veh_num),
                    fill_location='fill_loc %s %d' % (veh_str, i),
                    cost_per_gallon=(
                        Decimal('2.0') + (i * Decimal('0.1')) +
                        (veh_num * Decimal('0.01'))
                    ),
                    total_cost=(
                        (
                            Decimal('2.0') + (i * Decimal('0.1')) +
                            (veh_num * Decimal('0.01'))
                        ) * (1 + i)
                    ),
                    gallons=(i * 10) + veh_num,
                    reported_mpg=(20 + i) + (veh_num * 0.1),
                    notes='notes %s %d' % (veh_str, i)
                ))

    def _projects(self):
        # P1 - Active=77.77 Total=2546.89
        p1 = Project(name='P1', notes='ProjectOne')
        self.db.add(p1)
        self.db.add(BoMItem(
            project=p1,
            name='P1Item1',
            notes='P1Item1Notes',
            unit_cost=Decimal('11.11')
        ))
        self.db.add(BoMItem(
            project=p1,
            name='P1Item2',
            notes='P1Item2Notes',
            unit_cost=Decimal('22.22'),
            quantity=3,
            url='http://item2.p1.com'
        ))
        self.db.add(BoMItem(
            project=p1,
            name='P1Item3',
            notes='P1Item3Notes',
            unit_cost=Decimal('1234.56'),
            quantity=2,
            url='http://item3.p1.com',
            is_active=False
        ))
        p2 = Project(name='P2', notes='ProjectTwo')
        self.db.add(p2)
        # P3 - Active=3.0 Total=5.34
        p3 = Project(
            name='P3Inactive', notes='ProjectThreeInactive', is_active=False
        )
        self.db.add(p3)
        self.db.add(BoMItem(
            project=p3,
            name='P3Item2',
            notes='P3Item2Notes',
            unit_cost=Decimal('1.0'),
            quantity=3,
            url='http://item2.p3.com'
        ))
        self.db.add(BoMItem(
            project=p3,
            name='P3Item3',
            notes='P3Item3Notes',
            unit_cost=Decimal('2.34'),
            url='http://item3.p3.com',
            is_active=False
        ))



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""
import sys
from datetime import datetime, date
from decimal import Decimal

from ofxparse.ofxparse import Transaction
from pytz import UTC
from sqlalchemy.orm.query import Query
from sqlalchemy.sql.expression import null

from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.models.ofx_statement import OFXStatement
from biweeklybudget.models.account import Account
from biweeklybudget.tests.unit_helpers import binexp_to_dict

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import Mock, patch, call  # noqa
else:
    from unittest.mock import Mock, patch, call  # noqa

pbm = 'biweeklybudget.models.ofx_transaction'


class TestParamsFromOfxparserTransaction(object):

    def setup_method(self):
        trans = Transaction()
        trans.payee = 'PayeeName'
        trans.type = 'TType'
        trans.date = datetime(2017, 3, 10, 14, 15, 16)
        trans.amount = Decimal('123.45')
        trans.id = 'ABC123'
        trans.memo = 'TMemo'
        self.trans = trans
        self.stmt = Mock(spec_set=OFXStatement)
        self.acct_id = 2

    def test_simple(self):
        res = OFXTransaction.params_from_ofxparser_transaction(
            self.trans, self.acct_id, self.stmt
        )
        assert res == {
            'account_id': self.acct_id,
            'statement': self.stmt,
            'memo': 'TMemo',
            'name': 'PayeeName',
            'amount': Decimal('123.45'),
            'trans_type': 'TType',
            'date_posted': datetime(2017, 3, 10, 14, 15, 16, tzinfo=UTC),
            'fitid': 'ABC123',
            'sic': None,
            'mcc': ''
        }

    def test_cat_memo(self):
        res = OFXTransaction.params_from_ofxparser_transaction(
            self.trans, self.acct_id, self.stmt, cat_memo=True
        )
        assert res == {
            'account_id': self.acct_id,
            'statement': self.stmt,
            'name': 'PayeeNameTMemo',
            'amount': Decimal('123.45'),
            'trans_type': 'TType',
            'date_posted': datetime(2017, 3, 10, 14, 15, 16, tzinfo=UTC),
            'fitid': 'ABC123',
            'sic': None,
            'mcc': ''
        }

    def test_extra_attrs(self):
        self.trans.mcc = 'TMCC'
        self.trans.sic = 456
        self.trans.checknum = 789
        res = OFXTransaction.params_from_ofxparser_transaction(
            self.trans, self.acct_id, self.stmt
        )
        assert res == {
            'account_id': self.acct_id,
            'statement': self.stmt,
            'memo': 'TMemo',
            'name': 'PayeeName',
            'amount': Decimal('123.45'),
            'trans_type': 'TType',
            'date_posted': datetime(2017, 3, 10, 14, 15, 16, tzinfo=UTC),
            'fitid': 'ABC123',
            'sic': 456,
            'mcc': 'TMCC',
            'checknum': 789
        }

    def test_account_amount(self):
        ot = OFXTransaction(
            account=Mock(spec_set=Account, negate_ofx_amounts=False),
            amount=Decimal('123.45')
        )
        assert ot.account_amount == Decimal('123.45')

    def test_account_amount_negated(self):
        ot = OFXTransaction(
            account=Mock(spec_set=Account, negate_ofx_amounts=True),
            amount=Decimal('123.45')
        )
        assert ot.account_amount == Decimal('-123.45')

    @patch('%s.RECONCILE_BEGIN_DATE' % pbm, date(2017, 3, 17))
    def test_unreconciled(self):
        m_db = Mock()
        m_q = Mock(spec_set=Query)
        m_filt = Mock(spec_set=Query)
        m_db.query.return_value = m_q
        m_q.filter.return_value = m_filt
        res = OFXTransaction.unreconciled(m_db)
        assert res == m_filt
        assert len(m_db.mock_calls) == 2
        assert m_db.mock_calls[0] == call.query(OFXTransaction)
        kall = m_db.mock_calls[1]
        assert kall[0] == 'query().filter'
        expected1 = OFXTransaction.reconcile.__eq__(null())
        cutoff = datetime(2017, 3, 17, 0, 0, 0, tzinfo=UTC)
        expected2 = OFXTransaction.date_posted.__ge__(cutoff)
        expected3 = OFXTransaction.account.has(reconcile_trans=True)
        assert len(kall[1]) == 8
        assert str(expected1) == str(kall[1][0])
        assert binexp_to_dict(expected2) == binexp_to_dict(kall[1][1])
        assert str(kall[1][2]) == str(expected3)
        assert str(
            OFXTransaction.is_payment.__ne__(True)
        ) == str(kall[1][3])
        assert str(
            OFXTransaction.is_late_fee.__ne__(True)
        ) == str(kall[1][4])
        assert str(
            OFXTransaction.is_interest_charge.__ne__(True)
        ) == str(kall[1][5])
        assert str(
            OFXTransaction.is_other_fee.__ne__(True)
        ) == str(kall[1][6])
        assert str(
            OFXTransaction.is_interest_payment.__ne__(True)
        ) == str(kall[1][7])



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from datetime import datetime
import os
import argparse
import logging
from copy import deepcopy
from io import StringIO
import importlib
import json

from ofxparse import OfxParser
from biweeklybudget.vendored.ofxclient.account \
    import Account as OfxClientAccount

from biweeklybudget.vault import Vault
from biweeklybudget.cliutils import set_log_debug, set_log_info
from biweeklybudget.ofxapi import apiclient

logger = logging.getLogger(__name__)

# suppress requests logging
requests_log = logging.getLogger("requests")
requests_log.setLevel(logging.WARNING)
requests_log.propagate = True


class OfxGetter(object):

    @staticmethod
    def accounts(client):
        """
        Return a dict of account information of ofxgetter-enabled accounts,
        str account name to dict of information about the account.

        :param client: API client
        :type client: Instance of :py:class:`~.OfxApiLocal` or
          :py:class:`~.OfxApiRemote`
        :returns: dict of account information; see
          :py:meth:`~.OfxApiLocal.get_accounts` for details.
        :rtype: dict
        """
        return client.get_accounts()

    def __init__(self, client, savedir='./'):
        """
        Initialize OfxGetter class.

        :param client: API client
        :type client: Instance of :py:class:`~.OfxApiLocal` or
          :py:class:`~.OfxApiRemote`
        :param savedir: directory/path to save statements in
        :type savedir: str
        """
        self._client = client
        self.savedir = savedir
        self._account_data = self.accounts(self._client)
        logger.debug('Initialized with data for %d accounts',
                     len(self._account_data))
        self._accounts = {}
        self.vault = Vault()
        for acct_name in self._account_data.keys():
            data = self._account_data[acct_name]['config']
            vault_path = self._account_data[acct_name]['vault_path']
            logger.debug('Getting secrets for account %s', acct_name)
            secrets = self.vault.read(vault_path)
            if list(data.keys()) == ['key']:
                logger.debug(
                    'Account %s has ofxgetter_config_json stored in Vault key: '
                    '%s' % (acct_name, data['key'])
                )
                if data['key'] not in secrets:
                    raise RuntimeError(
                        'Account %s should have ofxgetter_config_json stored '
                        'in Vault key %s, but Vault entry at %s has no such '
                        'key' % (
                            acct_name, data['key'], vault_path
                        )
                    )
                data = json.loads(secrets[data['key']])
            data['institution']['password'] = secrets['password']
            data['institution']['username'] = secrets['username']
            self._account_data[acct_name]['config'] = data
            if 'class_name' not in data:
                self._accounts[acct_name] = OfxClientAccount.deserialize(data)
        logger.debug('Initialized %d accounts', len(self._accounts))
        self.now_str = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')

    def get_ofx(self, account_name, write_to_file=True, days=30):
        """
        Download OFX from the specified account. Return it as a string.

        :param account_name: account name to download
        :type account_name: str
        :param write_to_file: if True, also write to a file named
          "<account_name>_<date stamp>.ofx"
        :type write_to_file: bool
        :param days: number of days of data to download
        :type days: int
        :return: OFX string
        :rtype: str
        """
        fname = None
        logger.debug('Downloading OFX for account: %s', account_name)
        if 'class_name' in self._account_data[account_name]['config']:
            ofxdata = self._get_ofx_scraper(account_name, days=days)
        else:
            acct = self._accounts[account_name]
            if logger.getEffectiveLevel() != logging.DEBUG:
                logger.debug(
                    'Disabling logging for ofxclient, which has bad logging'
                )
                oldlvl = logging.getLogger().getEffectiveLevel()
                logging.getLogger().setLevel(logging.WARNING)
            ofxdata = acct.download(days=days).read()
            if logger.getEffectiveLevel() != logging.DEBUG:
                logging.getLogger().setLevel(oldlvl)
                logger.debug('Re-enabling ofxclient logging')
        if write_to_file:
            fname = self._write_ofx_file(account_name, ofxdata)
        self._ofx_to_db(account_name, fname, ofxdata)
        return ofxdata

    def _ofx_to_db(self, account_name, fname, ofxdata):
        """
        Put OFX Data to the DB

        :param account_name: account name to download
        :type account_name: str
        :param ofxdata: raw OFX data
        :type ofxdata: str
        :param fname: filename OFX was written to
        :type fname: str
        """
        logger.debug('Parsing OFX')
        ofx = OfxParser.parse(StringIO(ofxdata))
        logger.debug('Updating OFX in DB')
        _, count_new, count_upd = self._client.update_statement_ofx(
            self._account_data[account_name]['id'], ofx, filename=fname
        )
        logger.info('Account "%s" - inserted %d new OFXTransaction(s), updated '
                    '%d existing OFXTransaction(s)',
                    account_name, count_new, count_upd)
        logger.debug('Done updating OFX in DB')

    def _get_ofx_scraper(self, account_name, days=30):
        """
        Get OFX via a ScreenScraper subclass.

        :param account_name: account name
        :type account_name: str
        :param days: number of days of data to download
        :type days: int
        :return: OFX string
        :rtype: str
        """
        data = self._account_data[account_name]['config']
        clsname = data['class_name']
        modname = data['module_name']
        logger.debug('Scraper - getting class %s from module %s',
                     clsname, modname)
        cls = getattr(
            importlib.import_module(modname),
            clsname
        )
        logger.debug('Getting secrets for account %s', account_name)
        secrets = self.vault.read(
            self._account_data[account_name]['vault_path']
        )
        if 'kwargs' in data:
            kwargs = deepcopy(data['kwargs'])
        else:
            kwargs = {}
        kwargs['username'] = secrets['username']
        kwargs['password'] = secrets['password']
        kwargs['savedir'] = os.path.join(self.savedir, account_name)
        acct = cls(**kwargs)
        ofxdata = acct.run()
        return ofxdata

    def _write_ofx_file(self, account_name, ofxdata):
        """
        Write OFX data to a file.

        :param account_name: account name
        :type account_name: str
        :param ofxdata: raw OFX data string
        :type ofxdata: str
        :returns: name of the file that was written
        :rtype: str
        """
        if not os.path.exists(os.path.join(self.savedir, account_name)):
            os.makedirs(os.path.join(self.savedir, account_name))
        fname = '%s_%s.ofx' % (account_name, self.now_str)
        fpath = os.path.join(self.savedir, account_name, fname)
        logger.debug('Writing %d bytes of OFX to: %s', len(ofxdata), fpath)
        with open(fpath, 'w') as fh:
            fh.write(ofxdata)
        logger.debug('Wrote OFX data to: %s', fpath)
        return fname


def parse_args():
    p = argparse.ArgumentParser(description='Download OFX transactions')
    p.add_argument('-v', '--verbose', dest='verbose', action='count', default=0,
                   help='verbose output. specify twice for debug-level output.')
    p.add_argument('-l', '--list-accts', dest='list', action='store_true',
                   help='list accounts and exit')
    p.add_argument('-i', '--institution', dest='institution',
                   action='store_true', default=False,
                   help='list all accounts for institution and exit')
    p.add_argument('-r', '--remote', dest='remote', action='store', type=str,
                   default=None,
                   help='biweeklybudget API URL to use instead of direct DB '
                        'access')
    p.add_argument('--ca-bundle', dest='ca_bundle', action='store', type=str,
                   default=None,
                   help='Path to CA certificate bundle file or directory to '
                        'use for SSL verification')
    p.add_argument('--client-cert', dest='client_cert', action='store',
                   type=str, default=None,
                   help='path to client certificate to use for SSL client '
                        'cert auth')
    p.add_argument('--client-key', dest='client_key', action='store',
                   type=str, default=None,
                   help='path to unencrypted client key to use for SSL client '
                        'cert auth, if key is not contained in the cert file')
    p.add_argument('-s', '--save-path', dest='save_path', action='store',
                   type=str, default=None,
                   help='Statement save path; must be specified when running '
                        'in remote (-r) mode.')
    p.add_argument('-d', '--days', dest='days', action='store', type=int,
                   default=30,
                   help='number of days of history to get; default 30')
    p.add_argument('ACCOUNT_NAME', type=str, action='store', default=None,
                   nargs='?',
                   help='Account name; omit to download all accounts')
    args = p.parse_args()
    return args


def main():
    global logger
    format = "[%(asctime)s %(levelname)s] %(message)s"
    logging.basicConfig(level=logging.WARNING, format=format)
    logger = logging.getLogger()

    args = parse_args()

    # set logging level
    if args.verbose > 1:
        set_log_debug(logger)
    elif args.verbose == 1:
        set_log_info(logger)
    if args.verbose <= 1:
        # if we're not in verbose mode, suppress routine logging for cron
        lgr = logging.getLogger('alembic')
        lgr.setLevel(logging.WARNING)
        lgr = logging.getLogger('biweeklybudget.db')
        lgr.setLevel(logging.WARNING)

    client = apiclient(
        api_url=args.remote, ca_bundle=args.ca_bundle,
        client_cert=args.client_cert, client_key=args.client_key
    )
    if args.list:
        for k in sorted(OfxGetter.accounts(client).keys()):
            print(k)
        raise SystemExit(0)

    if args.remote is None:
        from biweeklybudget import settings
        save_path = settings.STATEMENTS_SAVE_PATH
    else:
        if args.save_path is None:
            logger.error('ERROR: -s|--save-path must be specified when running '
                         'in remote mode.')
            raise SystemExit(1)
        save_path = os.path.abspath(args.save_path)

    getter = OfxGetter(client, save_path)

    if args.institution:
        if args.ACCOUNT_NAME is None:
            raise SystemExit('ERROR: Account name must be specified')
        getter = OfxGetter(client, save_path)
        acct = getter._accounts[args.ACCOUNT_NAME]
        logger.info('Authenticating to institution:')
        logger.debug(acct.institution.authenticate())
        logger.info('Getting list of accounts...')
        accts = acct.institution.accounts()
        print('Found %d accounts' % len(accts))
        for a in accts:
            print(a.serialize())
        raise SystemExit(0)

    if args.ACCOUNT_NAME is not None:
        getter.get_ofx(args.ACCOUNT_NAME, days=args.days)
        raise SystemExit(0)
    # else all of them
    total = 0
    success = 0
    for acctname in sorted(OfxGetter.accounts(client).keys()):
        try:
            total += 1
            getter.get_ofx(acctname, days=args.days)
            success += 1
        except Exception:
            logger.error(
                'Failed to download account %s', acctname, exc_info=True
            )
    if success != total:
        logger.warning('Downloaded %d of %d accounts', success, total)
        raise SystemExit(1)
    raise SystemExit(0)


if __name__ == "__main__":
    main()



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from flask.views import MethodView
from flask import render_template, jsonify, request
from datatables import DataTable
from sqlalchemy import or_, asc
from decimal import Decimal, ROUND_FLOOR
from datetime import datetime
from copy import copy

from biweeklybudget.flaskapp.app import app
from biweeklybudget.db import db_session
from biweeklybudget.models.fuel import FuelFill, Vehicle
from biweeklybudget.models.account import Account
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.models.transaction import Transaction
from biweeklybudget.flaskapp.views.searchableajaxview import SearchableAjaxView
from biweeklybudget.flaskapp.views.formhandlerview import FormHandlerView


logger = logging.getLogger(__name__)


class FuelView(MethodView):
    """
    Render the GET /fuel view using the ``fuel.html`` template.
    """

    def get(self):
        accts = {a.name: a.id for a in db_session.query(Account).all()}
        budgets = {}
        for b in db_session.query(Budget).all():
            if b.is_income:
                budgets['%s (income)' % b.name] = b.id
            else:
                budgets[b.name] = b.id
        vehicles = {
            v.id: {'name': v.name, 'is_active': v.is_active}
            for v in db_session.query(Vehicle).all()
        }
        return render_template(
            'fuel.html',
            accts=accts,
            budgets=budgets,
            vehicles=vehicles
        )


class FuelAjax(SearchableAjaxView):
    """
    Handle GET /ajax/fuelLog endpoint.
    """

    def _filterhack(self, qs, s, args):
        """
        DataTables 1.10.12 has built-in support for filtering based on a value
        in a specific column; when this is done, the filter value is set in
        ``columns[N][search][value]`` where N is the column number. However,
        the python datatables package used here only supports the global
        ``search[value]`` input, not the per-column one.

        However, the DataTable search is implemented by passing a callable
        to ``table.searchable()`` which takes two arguments, the current Query
        that's being built, and the user's ``search[value]`` input; this must
        then return a Query object with the search applied.

        In python datatables 0.4.9, this code path is triggered on
        ``if callable(self.search_func) and search.get("value", None):``

        As such, we can "trick" the table to use per-column searching (currently
        only if global searching is not being used) by examining the per-column
        search values in the request, and setting the search function to one
        (this method) that uses those values instead of the global
        ``search[value]``.

        :param qs: Query currently being built
        :type qs: ``sqlalchemy.orm.query.Query``
        :param s: user search value
        :type s: str
        :param args: args
        :type args: dict
        :return: Query with searching applied
        :rtype: ``sqlalchemy.orm.query.Query``
        """
        # Ok, build our filter...
        veh_filter = args['columns'][1]['search']['value']
        if veh_filter != '' and veh_filter != 'None':
            qs = qs.filter(FuelFill.vehicle_id == veh_filter)
        # search
        if s != '' and s != 'FILTERHACK':
            if len(s) < 3:
                return qs
            s = '%' + s + '%'
            qs = qs.filter(or_(
                FuelFill.notes.like(s)),
                FuelFill.fill_location.like(s)
            )
        return qs

    def get(self):
        """
        Render and return JSON response for GET /ajax/ofx
        """
        args = request.args.to_dict()
        args_dict = self._args_dict(args)
        if self._have_column_search(args_dict) and args['search[value]'] == '':
            args['search[value]'] = 'FILTERHACK'
        table = DataTable(
            args,
            FuelFill,
            db_session.query(FuelFill).filter(
                FuelFill.vehicle.has(is_active=True)
            ),
            [
                (
                    'date',
                    lambda i: i.date.strftime('%Y-%m-%d')
                ),
                (
                    'vehicle',
                    'vehicle.name'
                ),
                'odometer_miles',
                'reported_miles',
                'calculated_miles',
                'level_before',
                'level_after',
                'fill_location',
                'cost_per_gallon',
                'total_cost',
                (
                    'gallons',
                    lambda i: i.gallons.quantize(
                        Decimal('.001'), rounding=ROUND_FLOOR
                    )
                ),
                'reported_mpg',
                'calculated_mpg',
                'notes'
            ]
        )
        table.add_data(
            vehicle_id=lambda o: o.vehicle_id
        )
        if args['search[value]'] != '':
            table.searchable(lambda qs, s: self._filterhack(qs, s, args_dict))
        return jsonify(table.json())


class VehicleAjax(MethodView):
    """
    Handle GET /ajax/vehicle/<int:vehicle_id> endpoint.
    """

    def get(self, vehicle_id):
        v = db_session.query(Vehicle).get(vehicle_id)
        return jsonify(v.as_dict)


class VehicleFormHandler(FormHandlerView):
    """
    Handle POST /forms/vehicle
    """

    def validate(self, data):
        """
        Validate the form data. Return None if it is valid, or else a hash of
        field names to list of error strings for each field.

        :param data: submitted form data
        :type data: dict
        :return: None if no errors, or hash of field name to errors for that
          field
        """
        logger.warning(data)
        have_errors = False
        errors = {k: [] for k in data.keys()}
        if data.get('name', '').strip() == '':
            errors['name'].append('Name cannot be empty')
            have_errors = True
        if data.get('id', '').strip() == '':
            v = db_session.query(Vehicle).filter(
                Vehicle.name.__eq__(data.get('name'))
            ).all()
            if len(v) > 0:
                errors['name'].append('Name must be unique')
                have_errors = True
        else:
            v = db_session.query(Vehicle).filter(
                Vehicle.name.__eq__(data.get('name')),
                Vehicle.id.__ne__(int(data.get('id')))
            ).all()
            if len(v) > 0:
                errors['name'].append('Name must be unique')
                have_errors = True
        if have_errors:
            return errors
        return None

    def submit(self, data):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :param data: submitted form data
        :type data: dict
        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        if 'id' in data and data['id'].strip() != '':
            # updating an existing Vehicle
            veh = db_session.query(Vehicle).get(int(data['id']))
            if veh is None:
                raise RuntimeError("Error: no Vehicle with ID %s" % data['id'])
            action = 'updating Vehicle ' + data['id']
        else:
            veh = Vehicle()
            action = 'creating new Vehicle'
        veh.name = data['name'].strip()
        veh.is_active = data['is_active']
        logger.info('%s: %s', action, veh.as_dict)
        db_session.add(veh)
        db_session.commit()
        return 'Successfully saved Vehicle %d in database.' % veh.id


class FuelLogFormHandler(FormHandlerView):
    """
    Handle POST /forms/fuel
    """

    def validate(self, data):
        """
        Validate the form data. Return None if it is valid, or else a hash of
        field names to list of error strings for each field.

        :param data: submitted form data
        :type data: dict
        :return: None if no errors, or hash of field name to errors for that
          field
        """
        errors = {k: [] for k in data.keys()}
        errors = self._validate_date_ymd('date', data, errors)
        if data['add_trans'] is True:
            if data['account'] == 'None':
                errors['account'].append('Transactions must have an account')
            if data['budget'] == 'None':
                errors['budget'].append('Transactions must have a budget')
        if db_session.query(Vehicle).get(int(data['vehicle'])) is None:
            errors['vehicle'].append('Invalid Vehicle ID %s' % data['vehicle'])
        errors = self._validate_int('odometer_miles', data, errors)
        errors = self._validate_int('reported_miles', data, errors)
        errors = self._validate_not_empty('fill_location', data, errors)
        errors = self._validate_float('cost_per_gallon', data, errors)
        errors = self._validate_float('total_cost', data, errors)
        errors = self._validate_float('gallons', data, errors)
        errors = self._validate_float('reported_mpg', data, errors)
        errors = self._validate_int('vehicle', data, errors)
        errors = self._validate_int('level_before', data, errors)
        errors = self._validate_int('level_after', data, errors)
        if len(errors['total_cost']) == 0 and float(data['total_cost']) == 0:
            errors['total_cost'].append('Total Cost cannot be zero')
        for v in errors.values():
            if len(v) > 0:
                return errors
        return None

    def submit(self, data):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :param data: submitted form data
        :type data: dict
        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        total = Decimal(data['total_cost'])
        dt = datetime.strptime(data['date'], '%Y-%m-%d').date()
        veh_name = db_session.query(Vehicle).get(int(data['vehicle'])).name
        fill = FuelFill()
        fill.date = dt
        fill.vehicle_id = int(data['vehicle'])
        fill.odometer_miles = int(data['odometer_miles'])
        fill.reported_miles = int(data['reported_miles'])
        fill.level_before = int(data['level_before'])
        fill.level_after = int(data['level_after'])
        fill.fill_location = data['fill_location'].strip()
        fill.cost_per_gallon = Decimal(data['cost_per_gallon'])
        fill.total_cost = total
        fill.gallons = Decimal(data['gallons'])
        fill.reported_mpg = Decimal(data['reported_mpg'])
        fill.notes = data['notes'].strip()
        logger.info('Creating new FuelFill: %s', fill.as_dict)
        db_session.add(fill)
        db_session.commit()
        fill.calculate_mpg()
        db_session.commit()
        if data['add_trans'] is not True:
            return {
                'success_message': 'Successfully saved FuelFill %d '
                                   'in database.' % fill.id,
                'success': True,
                'fill_id': fill.id,
                'calculated_mpg': fill.calculated_mpg
            }
        budg = db_session.query(Budget).get(int(data['budget']))
        trans = Transaction(
            description='%s - FuelFill #%d (%s)' % (
                data['fill_location'].strip(), fill.id, veh_name
            ),
            date=dt,
            notes=data['notes'].strip(),
            budget_amounts={budg: total},
            account_id=int(data['account'])
        )
        logger.info('Creating new Transaction for FuelFill: %s', trans.as_dict)
        db_session.add(trans)
        db_session.commit()
        return {
            'success_message': 'Successfully saved FuelFill %d and '
                               ' Transaction %d in database.' % (
                                   fill.id, trans.id
                               ),
            'success': True,
            'trans_id': trans.id,
            'fill_id': fill.id,
            'calculated_mpg': fill.calculated_mpg
        }


class FuelMPGChartView(MethodView):
    """
    Handle GET /ajax/chart-data/fuel-economy endpoint.
    """

    def get(self):
        vehicles = {
            x.id: x.name for x in
            db_session.query(
                Vehicle
            ).filter(Vehicle.is_active.__eq__(True)).all()
        }
        veh_names = vehicles.values()
        datedict = {x: None for x in veh_names}
        data = {}
        mpgs = db_session.query(
            FuelFill
        ).filter(
            FuelFill.vehicle.has(is_active=True),
            FuelFill.calculated_mpg.__ne__(None)
        ).order_by(asc(FuelFill.date), asc(FuelFill.odometer_miles))
        for mpg in mpgs.all():
            ds = mpg.date.strftime('%Y-%m-%d')
            if ds not in data:
                data[ds] = copy(datedict)
                data[ds]['date'] = ds
            if mpg.calculated_mpg is None:
                data[ds][mpg.vehicle.name] = 0.0
            else:
                data[ds][mpg.vehicle.name] = Decimal(mpg.calculated_mpg)
        resdata = []
        last = None
        for k in sorted(data.keys()):
            if last is None:
                last = data[k]
                continue
            d = copy(data[k])
            for subk in veh_names:
                if d[subk] is None:
                    d[subk] = last[subk]
            last = d
            resdata.append(d)
        res = {
            'data': resdata,
            'keys': sorted(veh_names)
        }
        return jsonify(res)


class FuelPriceChartView(MethodView):
    """
    Handle GET /ajax/chart-data/fuel-prices endpoint.
    """

    def get(self):
        resdata = []
        prices = db_session.query(
            FuelFill
        ).filter(
            FuelFill.cost_per_gallon.__ne__(None)
        ).order_by(asc(FuelFill.date))
        for point in prices.all():
            ds = point.date.strftime('%Y-%m-%d')
            resdata.append({
                'date': ds,
                'price': float(point.cost_per_gallon)
            })
        res = {
            'data': resdata
        }
        return jsonify(res)


app.add_url_rule('/fuel', view_func=FuelView.as_view('fuel_view'))
app.add_url_rule(
    '/ajax/fuelLog',
    view_func=FuelAjax.as_view('fuel_ajax')
)
app.add_url_rule(
    '/ajax/vehicle/<int:vehicle_id>',
    view_func=VehicleAjax.as_view('vehicle_ajax')
)
app.add_url_rule(
    '/forms/vehicle',
    view_func=VehicleFormHandler.as_view('vehicle_form')
)
app.add_url_rule(
    '/forms/fuel',
    view_func=FuelLogFormHandler.as_view('fuel_form')
)
app.add_url_rule(
    '/ajax/chart-data/fuel-economy',
    view_func=FuelMPGChartView.as_view('fuel_mpg_chart_view')
)
app.add_url_rule(
    '/ajax/chart-data/fuel-prices',
    view_func=FuelPriceChartView.as_view('fuel_price_chart_view')
)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from flask.views import MethodView


class SearchableAjaxView(MethodView):
    """
    MethodView with helper methods for searching via DataTables ajax.
    """

    def _args_dict(self, args):
        """
        Given a 1-dimensional dict of request parameters like those used by
        DataTables (i.e. keys like ``columns[2][search][value]``), return a
        multidimensional dict representation of the same.

        :return: deep/nested dict
        :rtype: dict
        """
        d = {}
        for k, v in args.items():
            v = self._args_set_type(v)
            if '[' not in k:
                d[self._args_set_type(k)] = v
                continue
            k = k.replace('[', '|').replace(']', '|').replace('||', '|')
            k = k.strip('|')
            parts = k.split('|')
            ptr = d
            while len(parts) > 1:
                subk = self._args_set_type(parts.pop(0))
                if subk not in ptr:
                    ptr[subk] = {}
                ptr = ptr[subk]
            ptr[self._args_set_type(parts[0])] = v
        return d

    def _args_set_type(self, a):
        """
        Given a string portion of something in the argument dict, return it
        as the correct type.

        :param a: args dict key or value
        :type a: str
        :return: a in the proper type
        """
        try:
            if '%d' % int(a) == a:
                return int(a)
        except Exception:
            pass
        if a == 'true':
            return True
        if a == 'false':
            return False
        return a

    def _filterhack(self, qs, s, args):
        """
        DataTables 1.10.12 has built-in support for filtering based on a value
        in a specific column; when this is done, the filter value is set in
        ``columns[N][search][value]`` where N is the column number. However,
        the python datatables package used here only supports the global
        ``search[value]`` input, not the per-column one.

        However, the DataTable search is implemented by passing a callable
        to ``table.searchable()`` which takes two arguments, the current Query
        that's being built, and the user's ``search[value]`` input; this must
        then return a Query object with the search applied.

        In python datatables 0.4.9, this code path is triggered on
        ``if callable(self.search_func) and search.get("value", None):``

        As such, we can "trick" the table to use per-column searching (currently
        only if global searching is not being used) by examining the per-column
        search values in the request, and setting the search function to one
        (this method) that uses those values instead of the global
        ``search[value]``.

        :param qs: Query currently being built
        :type qs: ``sqlalchemy.orm.query.Query``
        :param s: user search value
        :type s: str
        :param args: args
        :type args: dict
        :return: Query with searching applied
        :rtype: ``sqlalchemy.orm.query.Query``
        """
        raise NotImplementedError()

    def _have_column_search(self, args):
        """
        Determine if we have a column filter/search in effect, and if so,
        should use :py:meth:`~._filterhack` as our search function.

        :param args: current request arguments
        :type args: dict
        :return: whether or not request asks for column filtering
        :rtype: bool
        """
        for col in args['columns']:
            if args['columns'][col]['search']['value'] != '':
                return True
        return False

    def get(self):
        """
        Render and return JSON response for GET.
        """
        raise NotImplementedError()
